"""
Training a LeNet with Monte-Carlo Dropout
=========================================

In this tutorial, we will train a LeNet classifier on the MNIST dataset using Monte-Carlo Dropout (MC Dropout), a computationally efficient Bayesian approximation method. To estimate the predictive mean and uncertainty (variance), we perform multiple forward passes through the network with dropout layers enabled in ``train`` mode.

For more information on Monte-Carlo Dropout, we refer the reader to the following resources:

- Dropout as a Bayesian Approximation: Representing Model Uncertainty in Deep Learning `ICML 2016 <https://browse.arxiv.org/pdf/1506.02142.pdf>`_
- What Uncertainties Do We Need in Bayesian Deep Learning for Computer Vision? `NeurIPS 2017 <https://browse.arxiv.org/pdf/1703.04977.pdf>`_

Training a LeNet with MC Dropout using TorchUncertainty models and PyTorch Lightning
-------------------------------------------------------------------------------------

In this part, we train a LeNet with dropout layers, based on the model and routines already implemented in TU.

1. Loading the utilities
~~~~~~~~~~~~~~~~~~~~~~~~

First, we have to load the following utilities from TorchUncertainty:

- the TUTrainer from TorchUncertainty utils
- the datamodule handling dataloaders: MNISTDataModule from torch_uncertainty.datamodules
- the model: lenet from torch_uncertainty.models
- the MC Dropout wrapper: mc_dropout, from torch_uncertainty.models.wrappers
- the classification training & evaluation routine in the torch_uncertainty.routines
- an optimization recipe in the torch_uncertainty.optim_recipes module.

We also need import the neural network utils within `torch.nn`.
"""
# %%
from pathlib import Path

from torch_uncertainty import TUTrainer
from torch import nn

from torch_uncertainty.datamodules import MNISTDataModule
from torch_uncertainty.models.lenet import lenet
from torch_uncertainty.models import mc_dropout
from torch_uncertainty.optim_recipes import optim_cifar10_resnet18
from torch_uncertainty.routines import ClassificationRoutine

# %%
# 2. Defining the Model and the Trainer
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In the following, we first create the trainer and instantiate 
# the datamodule that handles the MNIST dataset,
# dataloaders and transforms. We create the model using the
# blueprint from torch_uncertainty.models and we wrap it into an mc_dropout.
# To use the mc_dropout wrapper, **make sure that you use dropout modules** and
# not functionals. Moreover, **they have to be** instantiated in the __init__ method.

trainer = TUTrainer(accelerator="cpu", max_epochs=2, enable_progress_bar=False)

# datamodule
root = Path("data")
datamodule = MNISTDataModule(root=root, batch_size=128)


model = lenet(
    in_channels=datamodule.num_channels,
    num_classes=datamodule.num_classes,
    dropout_rate=0.4,
)

mc_model = mc_dropout(model, num_estimators=16, last_layer=False)

# %%
# 3. The Loss and the Training Routine
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# This is a classification problem, and we use CrossEntropyLoss as the (negative-log-)likelihood.
# We define the training routine using the classification training routine from
# torch_uncertainty.routines. We provide the number of classes
# the optimization recipe and tell the routine that our model is an ensemble at evaluation time.

routine = ClassificationRoutine(
    num_classes=datamodule.num_classes,
    model=mc_model,
    loss=nn.CrossEntropyLoss(),
    optim_recipe=optim_cifar10_resnet18(mc_model),
    is_ensemble=True,
)

# %%
# 4. Gathering Everything and Training the Model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We can now train the model using the trainer. We pass the routine and the datamodule
# to the fit and test methods of the trainer. It will automatically evaluate some uncertainty
# metrics that you will find in the table below.

trainer.fit(model=routine, datamodule=datamodule)
results = trainer.test(model=routine, datamodule=datamodule)

# %%
# 5. Testing the Model
# ~~~~~~~~~~~~~~~~~~~~
# Now that the model is trained, let's test it on MNIST. Don't forget to call
# .eval() to enable dropout at evaluation and get multiple (here 16) predictions.

import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision


def imshow(img):
    npimg = img.numpy()
    plt.imshow(np.transpose(npimg, (1, 2, 0)))
    plt.axis("off")
    plt.tight_layout()
    plt.show()


dataiter = iter(datamodule.val_dataloader())
images, labels = next(dataiter)

# print images
imshow(torchvision.utils.make_grid(images[:6, ...], padding=0))
print("Ground truth labels: ", " ".join(f"{labels[j]}" for j in range(6)))

routine.eval()
logits = routine(images).reshape(16, 128, 10)

probs = torch.nn.functional.softmax(logits, dim=-1)


for j in range(6):
    values, predicted = torch.max(probs[:, j], 1)
    print(
        f"MC-Dropout predictions for the image {j+1}: ",
        " ".join([str(image_id.item()) for image_id in predicted]),
    )

# %%
# Most of the time, we see that there is some disagreement between the samples of the dropout
# approximation of the posterior distribution.



"""
Improve Top-label Calibration with Temperature Scaling
======================================================

In this tutorial, we use *TorchUncertainty* to improve the calibration
of the top-label predictions and the reliability of the underlying neural network.

This tutorial provides extensive details on how to use the TemperatureScaler
class, however, this is done automatically in the classification routine when setting
the `calibration_set` to val or test.

Through this tutorial, we also see how to use the datamodules outside any Lightning trainers,
and how to use TorchUncertainty's models.

1. Loading the Utilities
~~~~~~~~~~~~~~~~~~~~~~~~

In this tutorial, we will need:

- TorchUncertainty's Calibration Error metric to compute to evaluate the top-label calibration with ECE and plot the reliability diagrams
- the CIFAR-100 datamodule to handle the data
- a ResNet 18 as starting model
- the temperature scaler to improve the top-label calibration
- a utility function to download HF models easily

If you use the classification routine, the plots will be automatically available in the tensorboard logs if you use the `log_plots` flag.
"""
# %%
from torch_uncertainty.datamodules import CIFAR100DataModule
from torch_uncertainty.metrics import CalibrationError
from torch_uncertainty.models.resnet import resnet
from torch_uncertainty.post_processing import TemperatureScaler
from torch_uncertainty.utils import load_hf

# %%
# 2. Loading a model from TorchUncertainty's HF
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To avoid training a model on CIFAR-100 from scratch, we load a model from Hugging Face.
# This can be done in a one liner:

# Build the model
model = resnet(in_channels=3, num_classes=100, arch=18, style="cifar", conv_bias=False)

# Download the weights (the config is not used here)
weights, config = load_hf("resnet18_c100")

# Load the weights in the pre-built model
model.load_state_dict(weights)

# %%
# 3. Setting up the Datamodule and Dataloaders
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To get the dataloader from the datamodule, just call prepare_data, setup, and
# extract the first element of the test dataloader list. There are more than one
# element if eval_ood is True: the dataloader of in-distribution data and the dataloader
# of out-of-distribution data. Otherwise, it is a list of 1 element.

dm = CIFAR100DataModule(root="./data", eval_ood=False, batch_size=32)
dm.prepare_data()
dm.setup("test")

# Get the full test dataloader (unused in this tutorial)
dataloader = dm.test_dataloader()[0]

# %%
# 4. Iterating on the Dataloader and Computing the ECE
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We first split the original test set into a calibration set and a test set for proper evaluation.
#
# When computing the ECE, you need to provide the likelihoods associated with the inputs.
# To do this, just call PyTorch's softmax.
#
# To avoid lengthy computations (without GPU), we restrict the calibration computation to a subset
# of the test set.

from torch.utils.data import DataLoader, random_split

# Split datasets
dataset = dm.test
cal_dataset, test_dataset, other = random_split(
    dataset, [1000, 1000, len(dataset) - 2000]
)
test_dataloader = DataLoader(test_dataset, batch_size=32)

# Initialize the ECE
ece = CalibrationError(task="multiclass", num_classes=100)

# Iterate on the calibration dataloader
for sample, target in test_dataloader:
    logits = model(sample)
    probs = logits.softmax(-1)
    ece.update(probs, target)

# Compute & print the calibration error
print(f"ECE before scaling - {ece.compute():.3%}.")

# %%
# We also compute and plot the top-label calibration figure. We see that the
# model is not well calibrated.
fig, ax = ece.plot()
fig.show()

# %%
# 5. Fitting the Scaler to Improve the Calibration
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The TemperatureScaler has one parameter that can be used to temper the softmax.
# We minimize the tempered cross-entropy on a calibration set that we define here as
# a subset of the test set and containing 1000 data. Look at the code run by TemperatureScaler
# `fit` method for more details.

# Fit the scaler on the calibration dataset
scaled_model = TemperatureScaler(model=model)
scaled_model.fit(calibration_set=cal_dataset)

# %%
# 6. Iterating Again to Compute the Improved ECE
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We can directly use the scaler as a calibrated model.
#
# Note that you will need to first reset the ECE metric to avoid mixing the scores of
# the previous and current iterations.

# Reset the ECE
ece.reset()

# Iterate on the test dataloader
for sample, target in test_dataloader:
    logits = scaled_model(sample)
    probs = logits.softmax(-1)
    ece.update(probs, target)

print(f"ECE after scaling - {ece.compute():.3%}.")

# %%
# We finally compute and plot the scaled top-label calibration figure. We see
# that the model is now better calibrated.
fig, ax = ece.plot()
fig.show()

# %%
# The top-label calibration should be improved.
#
# Notes
# ~~~~~
#
# Temperature scaling is very efficient when the calibration set is representative of the test set.
# In this case, we say that the calibration and test set are drawn from the same distribution.
# However, this may not hold true in real-world cases where dataset shift could happen.

# %%
# References
# ~~~~~~~~~~
#
# - **Expected Calibration Error:** Naeini, M. P., Cooper, G. F., & Hauskrecht, M. (2015). Obtaining Well Calibrated Probabilities Using Bayesian Binning. In `AAAI 2015 <https://arxiv.org/pdf/1411.0160.pdf>`_.
# - **Temperature Scaling:** Guo, C., Pleiss, G., Sun, Y., & Weinberger, K. Q. (2017). On calibration of modern neural networks. In `ICML 2017 <https://arxiv.org/pdf/1706.04599.pdf>`_.



from abc import ABC, abstractmethod

from torch import Tensor, nn
from torch.utils.data import Dataset


class PostProcessing(ABC, nn.Module):
    def __init__(self, model: nn.Module | None = None):
        super().__init__()
        self.model = model
        self.trained = False

    def set_model(self, model: nn.Module) -> None:
        self.model = model

    @abstractmethod
    def fit(self, dataset: Dataset) -> None:
        pass

    @abstractmethod
    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        pass



from torchvision.transforms import InterpolationMode


def interpolation_modes_from_str(val: str) -> InterpolationMode:
    val = val.lower()
    inverse_modes_mapping = {
        "nearest": InterpolationMode.NEAREST,
        "bilinear": InterpolationMode.BILINEAR,
        "bicubic": InterpolationMode.BICUBIC,
        "box": InterpolationMode.BOX,
        "hamming": InterpolationMode.HAMMING,
        "lanczos": InterpolationMode.LANCZOS,
    }
    return inverse_modes_mapping[val]



import logging
import os
import shutil
from collections.abc import Callable
from importlib import util
from pathlib import Path
from typing import Literal, NamedTuple

from huggingface_hub import hf_hub_download
from PIL import Image

if util.find_spec("cv2"):
    import cv2

    cv2_installed = True
else:  # coverage: ignore
    cv2_installed = False
import numpy as np
from torchvision import tv_tensors
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    download_and_extract_archive,
)


class MUADClass(NamedTuple):
    name: str
    id: int
    color: tuple[int, int, int]


class MUAD(VisionDataset):
    classes_url = "https://raw.githubusercontent.com/torch-uncertainty/dataset-metadata/main/segmentation/muad/classes.json"
    classes_md5 = "1db6e6143939824792f0af11a4fe7bb1"  # avoid replacement attack
    base_url = "https://zenodo.org/records/10619959/files/"

    zip_md5 = {
        "train": "cea6a672225b10dda1add8b2974a5982",
        "train_depth": "934d122ac09e0471db62ae68c3456b0f",
        "val": "957af9c1c36f0a85c33279e06b6cf8d8",
        "val_depth": "0282030d281aeffee3335f713ba12373",
    }

    small_muad_url = "ENSTA-U2IS/miniMUAD"

    _num_samples = {
        "full": {
            "train": 3420,
            "val": 492,
            "test": ...,
        },
        "small": {
            "train": 400,
            "val": 54,
            "test": 112,
            "ood": 20,
        },
    }

    classes = [
        MUADClass("road", 0, (128, 64, 128)),
        MUADClass("sidewalk", 1, (244, 35, 232)),
        MUADClass("building", 2, (70, 70, 70)),
        MUADClass("wall", 3, (102, 102, 156)),
        MUADClass("fence", 4, (190, 153, 153)),
        MUADClass("pole", 5, (153, 153, 153)),
        MUADClass("traffic_light", 6, (250, 170, 30)),
        MUADClass("traffic_sign", 7, (220, 220, 0)),
        MUADClass("vegetation", 8, (107, 142, 35)),
        MUADClass("terrain", 9, (152, 251, 152)),
        MUADClass("sky", 10, (70, 130, 180)),
        MUADClass("person", 11, (220, 20, 60)),
        MUADClass("rider", 12, (255, 0, 0)),
        MUADClass("car", 13, (0, 0, 142)),
        MUADClass("truck", 14, (0, 0, 70)),
        MUADClass("bus", 15, (0, 60, 100)),
        MUADClass("train", 16, (0, 80, 100)),
        MUADClass("motorcycle", 17, (0, 0, 230)),
        MUADClass("bicycle", 18, (119, 11, 32)),
        MUADClass("bear deer cow", 19, (255, 228, 196)),
        MUADClass("garbage_bag stand_food trash_can", 20, (128, 128, 0)),
        MUADClass("unlabeled", 21, (0, 0, 0)),  # id 255 or 21
    ]

    targets: list[Path] = []

    def __init__(
        self,
        root: str | Path,
        split: Literal["train", "val", "test", "ood"],
        version: Literal["small", "full"] = "full",
        min_depth: float | None = None,
        max_depth: float | None = None,
        target_type: Literal["semantic", "depth"] = "semantic",
        transforms: Callable | None = None,
        download: bool = False,
    ) -> None:
        """The MUAD Dataset.

        Args:
            root (str): Root directory of dataset where directory 'leftImg8bit'
                and 'leftLabel' or 'leftDepth' are located.
            split (str, optional): The image split to use, 'train' or 'val'.
            version (str, optional): The version of the dataset to use, 'small'
                or 'full'. Defaults to 'full'.
            min_depth (float, optional): The maximum depth value to use if
                target_type is 'depth'. Defaults to None.
            max_depth (float, optional): The maximum depth value to use if
                target_type is 'depth'. Defaults to None.
            target_type (str, optional): The type of target to use, 'semantic'
                or 'depth'.
            transforms (callable, optional): A function/transform that takes in
                a tuple of PIL images and returns a transformed version.
            download (bool, optional): If true, downloads the dataset from the
                internet and puts it in root directory. If dataset is already
                downloaded, it is not downloaded again.

        Reference:
            https://muad-dataset.github.io

        Note:
            MUAD cannot be used for commercial purposes. Read MUAD's license
            carefully before using it and verify that you can comply.
        """
        if not cv2_installed:  # coverage: ignore
            raise ImportError(
                "The cv2 library is not installed. Please install"
                "torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )

        if version == "small" and target_type == "depth":
            raise ValueError("Depth target is not available for the small version of MUAD.")

        logging.info(
            "MUAD is restricted to non-commercial use. By using MUAD, you "
            "agree to the terms and conditions."
        )

        dataset_root = Path(root) / "MUAD" if version == "full" else Path(root) / "MUAD_small"

        super().__init__(dataset_root, transforms=transforms)
        self.min_depth = min_depth
        self.max_depth = max_depth

        if split not in ["train", "val", "test", "ood"]:
            raise ValueError(f"split must be one of ['train', 'val']. Got {split}.")
        self.split = split
        self.version = version
        self.target_type = target_type

        if not self.check_split_integrity("leftImg8bit"):
            if download:
                self._download(split=split)
            else:
                raise FileNotFoundError(
                    f"MUAD {split} split not found or incomplete. Set download=True to download it."
                )

        if not self.check_split_integrity("leftLabel") and target_type == "semantic":
            if download:
                self._download(split=split)
            else:
                raise FileNotFoundError(
                    f"MUAD {split} split not found or incomplete. Set download=True to download it."
                )

        if not self.check_split_integrity("leftDepth") and target_type == "depth":
            if download:
                self._download(split=f"{split}_depth")
                # Depth target for train are in a different folder
                # thus we move them to the correct folder
                if split == "train":
                    shutil.move(
                        self.root / f"{split}_depth",
                        self.root / split / "leftDepth",
                    )
            else:
                raise FileNotFoundError(
                    f"MUAD {split} split not found or incomplete. Set download=True to download it."
                )

        self._make_dataset(self.root / split)

    def __getitem__(self, index: int) -> tuple[tv_tensors.Image, tv_tensors.Mask]:
        """Get the sample at the given index.

        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is either a segmentation mask
                or a depth map.
        """
        image = tv_tensors.Image(Image.open(self.samples[index]).convert("RGB"))
        if self.target_type == "semantic":
            target = tv_tensors.Mask(Image.open(self.targets[index]))
        else:
            os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
            target = Image.fromarray(
                cv2.imread(
                    str(self.targets[index]),
                    cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH,
                )
            )
            # TODO: in the long run it would be better to use a custom
            # tv_tensor for depth maps (e.g. tv_tensors.DepthMap)
            target = np.asarray(target, np.float32)
            target = tv_tensors.Mask(400 * (1 - target))  # convert to meters
            target[(target <= self.min_depth) | (target > self.max_depth)] = float("nan")

        if self.transforms is not None:
            image, target = self.transforms(image, target)

        return image, target

    def check_split_integrity(self, folder: str) -> bool:
        split_path = self.root / self.split
        return (
            split_path.is_dir() and len(list((split_path / folder).glob("**/*"))) == self.__len__()
        )

    def __len__(self) -> int:
        """The number of samples in the dataset."""
        return self._num_samples[self.version][self.split]

    def _make_dataset(self, path: Path) -> None:
        """Create a list of samples and targets.

        Args:
            path (Path): The path to the dataset.
        """
        if "depth" in path.name:
            raise NotImplementedError(
                "Depth mode is not implemented yet. Raise an issue if you need it."
            )
        self.samples = sorted((path / "leftImg8bit/").glob("**/*"))
        if self.target_type == "semantic":
            self.targets = sorted((path / "leftLabel/").glob("**/*"))
        elif self.target_type == "depth":
            self.targets = sorted((path / "leftDepth/").glob("**/*"))
        else:
            raise ValueError(
                f"target_type must be one of ['semantic', 'depth']. Got {self.target_type}."
            )

    def _download(self, split: str) -> None:
        """Download and extract the chosen split of the dataset."""
        if self.version == "small":
            filename = f"{split}.zip"
            downloaded_file = hf_hub_download(
                repo_id=self.small_muad_url, filename=filename, repo_type="dataset"
            )
            shutil.unpack_archive(downloaded_file, extract_dir=self.root)
        else:
            split_url = self.base_url + split + ".zip"
            download_and_extract_archive(split_url, self.root, md5=self.zip_md5[split])

    @property
    def color_palette(self) -> np.ndarray:
        return [c.color for c in self.classes]



"""
Training a LeNet with Monte Carlo Batch Normalization
=====================================================

In this tutorial, we will train a LeNet classifier on the MNIST dataset using Monte-Carlo Batch Normalization (MCBN), a post-hoc Bayesian approximation method. 

Training a LeNet with MCBN using TorchUncertainty models and PyTorch Lightning
------------------------------------------------------------------------------
In this part, we train a LeNet with batch normalization layers, based on the model and routines already implemented in TU.

1. Loading the utilities
~~~~~~~~~~~~~~~~~~~~~~~~

First, we have to load the following utilities from TorchUncertainty:

- the TUTrainer from our framework
- the datamodule handling dataloaders: MNISTDataModule from torch_uncertainty.datamodules
- the model: LeNet, which lies in torch_uncertainty.models
- the MC Batch Normalization wrapper: mc_batch_norm, which lies in torch_uncertainty.post_processing
- the classification training routine in the torch_uncertainty.routines
- an optimization recipe in the torch_uncertainty.optim_recipes module.

We also need import the neural network utils within `torch.nn`.
"""
# %%
from pathlib import Path

from torch import nn

from torch_uncertainty import TUTrainer
from torch_uncertainty.datamodules import MNISTDataModule
from torch_uncertainty.models.lenet import lenet
from torch_uncertainty.optim_recipes import optim_cifar10_resnet18
from torch_uncertainty.post_processing.mc_batch_norm import MCBatchNorm
from torch_uncertainty.routines import ClassificationRoutine

# %%
# 2. Creating the necessary variables
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# In the following, we define the root of the datasets and the
# logs. We also create the datamodule that handles the MNIST dataset
# dataloaders and transforms.

trainer = TUTrainer(accelerator="cpu", max_epochs=2, enable_progress_bar=False)

# datamodule
root = Path("data")
datamodule = MNISTDataModule(root, batch_size=128)


model = lenet(
    in_channels=datamodule.num_channels,
    num_classes=datamodule.num_classes,
    norm=nn.BatchNorm2d,
)

# %%
# 3. The Loss and the Training Routine
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# This is a classification problem, and we use CrossEntropyLoss as likelihood.
# We define the training routine using the classification training routine from
# torch_uncertainty.training.classification. We provide the number of classes,
# and the optimization recipe.

routine = ClassificationRoutine(
    num_classes=datamodule.num_classes,
    model=model,
    loss=nn.CrossEntropyLoss(),
    optim_recipe=optim_cifar10_resnet18(model),
)

# %%
# 4. Gathering Everything and Training the Model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# You can also save the results in a variable by saving the output of
# `trainer.test`.

trainer.fit(model=routine, datamodule=datamodule)
perf = trainer.test(model=routine, datamodule=datamodule)

# %%
# 5. Wrapping the Model in a MCBatchNorm
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We can now wrap the model in a MCBatchNorm to add stochasticity to the
# predictions. We specify that the BatchNorm layers are to be converted to
# MCBatchNorm layers, and that we want to use 8 stochastic estimators.
# The amount of stochasticity is controlled by the ``mc_batch_size`` argument.
# The larger the ``mc_batch_size``, the more stochastic the predictions will be.
# The authors suggest 32 as a good value for ``mc_batch_size`` but we use 4 here
# to highlight the effect of stochasticity on the predictions.

routine.model = MCBatchNorm(
    routine.model, num_estimators=8, convert=True, mc_batch_size=16
)
routine.model.fit(datamodule.train)
routine = routine.eval()  # To avoid prints

# %%
# 6. Testing the Model
# ~~~~~~~~~~~~~~~~~~~~
# Now that the model is trained, let's test it on MNIST. Don't forget to call
# .eval() to enable Monte Carlo batch normalization at evaluation (sometimes called inference).
# In this tutorial, we plot the most uncertain images, i.e. the images for which
# the variance of the predictions is the highest.
# Please note that we apply a reshape to the logits to determine the dimension corresponding to the ensemble
# and to the batch. As for TorchUncertainty 2.0, the ensemble dimension is merged with the batch dimension
# in this order (num_estimator x batch, classes).

import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision


def imshow(img):
    npimg = img.numpy()
    plt.imshow(np.transpose(npimg, (1, 2, 0)))
    plt.axis("off")
    plt.tight_layout()
    plt.show()


dataiter = iter(datamodule.val_dataloader())
images, labels = next(dataiter)

routine.eval()
logits = routine(images).reshape(8, 128, 10)  # num_estimators, batch_size, num_classes

probs = torch.nn.functional.softmax(logits, dim=-1)
most_uncertain = sorted(probs.var(0).sum(-1).topk(4).indices)

# print images
imshow(torchvision.utils.make_grid(images[most_uncertain, ...]))
print("Ground truth: ", " ".join(f"{labels[j]}" for j in range(4)))

for j in most_uncertain:
    values, predicted = torch.max(probs[:, j], 1)
    print(
        f"Predicted digits for the image {j}: ",
        " ".join([str(image_id.item()) for image_id in predicted]),
    )

# %%
# The predictions are mostly erroneous, which is expected since we selected
# the most uncertain images. We also see that there stochasticity in the
# predictions, as the predictions for the same image differ depending on the
# stochastic estimator used.



# ruff: noqa: F401
from .bayes_conv import BayesConv1d, BayesConv2d, BayesConv3d
from .bayes_linear import BayesLinear
from .lpbnn import LPBNNConv2d, LPBNNLinear

bayesian_modules = (
    BayesConv1d,
    BayesConv2d,
    BayesConv3d,
    BayesLinear,
    LPBNNLinear,
    LPBNNConv2d,
)



import numpy as np
import torch
from torch import Tensor, distributions, nn


class TrainableDistribution(nn.Module):
    lsqrt2pi = torch.tensor(np.log(np.sqrt(2 * np.pi)))

    def __init__(
        self,
        mu: Tensor,
        rho: Tensor,
    ) -> None:
        super().__init__()
        self.mu = mu
        self.rho = rho
        self.sigma = None
        self.weight = None

    def sample(self) -> Tensor:
        w_sample = torch.randn(size=self.mu.shape, device=self.mu.device)
        self.sigma = torch.log1p(torch.exp(self.rho)).to(self.mu.device)
        self.weight = self.mu + self.sigma * w_sample
        return self.weight

    def log_posterior(self, weight: Tensor | None = None) -> Tensor:
        if self.weight is None or self.sigma is None:
            raise ValueError("Sample the weights before querying the log posterior.")

        if weight is None:  # coverage: ignore
            weight = self.weight

        lposterior = (
            self.lsqrt2pi
            + torch.log(self.sigma)
            + (((weight - self.mu) ** 2) / (2 * self.sigma**2))
            + 0.5
        )
        return -lposterior.sum()


class CenteredGaussianMixture(nn.Module):
    def __init__(
        self,
        sigma_1: float,
        sigma_2: float,
        pi: float,
    ) -> None:
        """Create a mixture of two centered Gaussian distributions.

        Args:
            sigma_1 (float): Standard deviation of the first Gaussian.
            sigma_2 (float): Standard deviation of the second Gaussian.
            pi (float): Mixing coefficient.
        """
        super().__init__()
        self.register_buffer("pi", torch.tensor([pi, 1 - pi]))
        self.register_buffer("mus", torch.zeros(2))
        self.register_buffer("sigmas", torch.tensor([sigma_1, sigma_2]))

    def log_prob(self, weight: Tensor) -> Tensor:
        mix = distributions.Categorical(self.pi)
        normals = distributions.Normal(self.mus, self.sigmas)
        distribution = distributions.MixtureSameFamily(mix, normals)
        return distribution.log_prob(weight).sum()



import copy
from pathlib import Path
from typing import Literal

import torchvision.transforms as T
import yaml
from timm.data.auto_augment import rand_augment_transform
from timm.data.mixup import Mixup
from torch import nn
from torch.utils.data import DataLoader, Subset
from torchvision.datasets import DTD, SVHN, ImageNet, INaturalist

from torch_uncertainty.datamodules import TUDataModule
from torch_uncertainty.datasets.classification import (
    ImageNetA,
    ImageNetC,
    ImageNetO,
    ImageNetR,
    OpenImageO,
)
from torch_uncertainty.utils import (
    create_train_val_split,
    interpolation_modes_from_str,
)


class ImageNetDataModule(TUDataModule):
    num_classes = 1000
    num_channels = 3
    test_datasets = ["r", "o", "a"]
    ood_datasets = [
        "inaturalist",
        "imagenet-o",
        "svhn",
        "textures",
        "openimage-o",
    ]
    training_task = "classification"
    mean = (0.485, 0.456, 0.406)
    std = (0.229, 0.224, 0.225)
    train_indices = None
    val_indices = None

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        eval_ood: bool = False,
        eval_shift: bool = False,
        shift_severity: int = 1,
        val_split: float | Path | None = None,
        ood_ds: str = "openimage-o",
        test_alt: str | None = None,
        procedure: str | None = None,
        train_size: int = 224,
        interpolation: str = "bilinear",
        basic_augment: bool = True,
        rand_augment_opt: str | None = None,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
    ) -> None:
        """DataModule for ImageNet.

        Args:
            root (str): Root directory of the datasets.
            eval_ood (bool): Whether to evaluate out-of-distribution
                performance. Defaults to ``False``.
            eval_shift (bool): Whether to evaluate on shifted data. Defaults to
                ``False``.
            shift_severity: int = 1,
            batch_size (int): Number of samples per batch.
            val_split (float or Path): Share of samples to use for validation
                or path to a yaml file containing a list of validation images
                ids. Defaults to ``0.0``.
            ood_ds (str): Which out-of-distribution dataset to use. Defaults to
                ``"openimage-o"``.
            test_alt (str): Which test set to use. Defaults to ``None``.
            procedure (str): Which procedure to use. Defaults to ``None``.
            train_size (int): Size of training images. Defaults to ``224``.
            interpolation (str): Interpolation method for the Resize Crops.
                Defaults to ``"bilinear"``.
            basic_augment (bool): Whether to apply base augmentations. Defaults to
                ``True``.
            rand_augment_opt (str): Which RandAugment to use. Defaults to ``None``.
            num_workers (int): Number of workers to use for data loading. Defaults
                to ``1``.
            pin_memory (bool): Whether to pin memory. Defaults to ``True``.
            persistent_workers (bool): Whether to use persistent workers. Defaults
                to ``True``.
        """
        super().__init__(
            root=Path(root),
            batch_size=batch_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.eval_ood = eval_ood
        self.eval_shift = eval_shift
        self.shift_severity = shift_severity
        if val_split and not isinstance(val_split, float):
            val_split = Path(val_split)
            self.train_indices, self.val_indices = read_indices(val_split)
        self.val_split = val_split
        self.ood_ds = ood_ds
        self.test_alt = test_alt
        self.interpolation = interpolation_modes_from_str(interpolation)

        if test_alt is None:
            self.dataset = ImageNet
        elif test_alt == "r":
            self.dataset = ImageNetR
        elif test_alt == "o":
            self.dataset = ImageNetO
        elif test_alt == "a":
            self.dataset = ImageNetA
        else:
            raise ValueError(f"The alternative {test_alt} is not known.")

        if ood_ds == "inaturalist":
            self.ood_dataset = INaturalist
        elif ood_ds == "imagenet-o":
            self.ood_dataset = ImageNetO
        elif ood_ds == "svhn":
            self.ood_dataset = SVHN
        elif ood_ds == "textures":
            self.ood_dataset = DTD
        elif ood_ds == "openimage-o":
            self.ood_dataset = OpenImageO
        else:
            raise ValueError(f"The dataset {ood_ds} is not supported.")
        self.shift_dataset = ImageNetC

        self.procedure = procedure

        if basic_augment:
            basic_transform = T.Compose(
                [
                    T.RandomResizedCrop(train_size, interpolation=self.interpolation),
                    T.RandomHorizontalFlip(),
                ]
            )
        else:
            basic_transform = nn.Identity()

        if self.procedure is None:
            if rand_augment_opt is not None:
                main_transform = rand_augment_transform(rand_augment_opt, {})
            else:
                main_transform = nn.Identity()
        elif self.procedure == "ViT":
            train_size = 224
            main_transform = T.Compose(
                [
                    Mixup(mixup_alpha=0.2, cutmix_alpha=1.0),
                    rand_augment_transform("rand-m9-n2-mstd0.5", {}),
                ]
            )
        elif self.procedure == "A3":
            train_size = 160
            main_transform = rand_augment_transform("rand-m6-mstd0.5-inc1", {})
        else:
            raise ValueError("The procedure is unknown")

        self.train_transform = T.Compose(
            [
                T.ToTensor(),
                basic_transform,
                main_transform,
                T.Normalize(mean=self.mean, std=self.std),
            ]
        )

        self.test_transform = T.Compose(
            [
                T.ToTensor(),
                T.Resize(256, interpolation=self.interpolation),
                T.CenterCrop(224),
                T.Normalize(mean=self.mean, std=self.std),
            ]
        )

    def _verify_splits(self, split: str) -> None:
        if split not in list(self.root.iterdir()):
            raise FileNotFoundError(
                f"a {split} Imagenet split was not found in {self.root},"
                f" make sure the folder contains a subfolder named {split}"
            )

    def prepare_data(self) -> None:  # coverage: ignore
        if self.test_alt is not None:
            self.data = self.dataset(
                self.root,
                split="val",
                download=True,
            )
        if self.eval_ood:
            if self.ood_ds == "inaturalist":
                self.ood = self.ood_dataset(
                    self.root,
                    version="2021_valid",
                    download=True,
                    transform=self.test_transform,
                )
            elif self.ood_ds != "textures":
                self.ood = self.ood_dataset(
                    self.root,
                    split="test",
                    download=True,
                    transform=self.test_transform,
                )
            else:
                self.ood = self.ood_dataset(
                    self.root,
                    split="train",
                    download=True,
                    transform=self.test_transform,
                )
        if self.eval_shift:
            self.shift_dataset(
                self.root,
                download=True,
                transform=self.test_transform,
                shift_severity=self.shift_severity,
            )

    def setup(self, stage: Literal["fit", "test"] | None = None) -> None:
        if stage == "fit" or stage is None:
            if self.test_alt is not None:
                raise ValueError("The test_alt argument is not supported for training.")
            full = self.dataset(
                self.root,
                split="train",
                transform=self.train_transform,
            )
            if self.val_split and isinstance(self.val_split, float):
                self.train, self.val = create_train_val_split(
                    full,
                    self.val_split,
                    self.test_transform,
                )
            elif isinstance(self.val_split, Path):
                self.train = Subset(full, self.train_indices)
                # TODO: improve the performance
                self.val = copy.deepcopy(Subset(full, self.val_indices))
                self.val.dataset.transform = self.test_transform
            else:
                self.train = full
                self.val = self.dataset(
                    self.root,
                    split="val",
                    transform=self.test_transform,
                )
        if stage == "test" or stage is None:
            self.test = self.dataset(
                self.root,
                split="val",
                transform=self.test_transform,
            )
        if stage not in ["fit", "test", None]:
            raise ValueError(f"Stage {stage} is not supported.")

        if self.eval_ood:
            if self.ood_ds == "inaturalist":
                self.ood = self.ood_dataset(
                    self.root,
                    version="2021_valid",
                    transform=self.test_transform,
                )
            else:
                self.ood = self.ood_dataset(
                    self.root,
                    transform=self.test_transform,
                    download=True,
                )

        if self.eval_shift:
            self.shift = self.shift_dataset(
                self.root,
                download=False,
                transform=self.test_transform,
                shift_severity=self.shift_severity,
            )

    def test_dataloader(self) -> list[DataLoader]:
        """Get the test dataloaders for ImageNet.

        Return:
            list[DataLoader]: ImageNet test set (in distribution data), OOD dataset test split
            (out-of-distribution data), and/or ImageNetC data.
        """
        dataloader = [self._data_loader(self.test)]
        if self.eval_ood:
            dataloader.append(self._data_loader(self.ood))
        if self.eval_shift:
            dataloader.append(self._data_loader(self.shift))
        return dataloader


def read_indices(path: Path) -> list[str]:  # coverage: ignore
    """Read a file and return its lines as a list.

    Args:
        path (Path): Path to the file.

    Returns:
        list[str]: list of filenames.
    """
    if not path.is_file():
        raise ValueError(f"{path} is not a file.")
    with path.open("r") as f:
        indices = yaml.safe_load(f)
        return indices["train"], indices["val"]



from pathlib import Path

import pytest
from torch import nn

from tests._dummies import (
    DummySegmentationBaseline,
    DummySegmentationDataModule,
)
from torch_uncertainty import TUTrainer
from torch_uncertainty.optim_recipes import optim_cifar10_resnet18
from torch_uncertainty.routines import SegmentationRoutine


class TestSegmentation:
    def test_one_estimator_two_classes(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        root = Path(__file__).parent.absolute().parents[0] / "data"
        dm = DummySegmentationDataModule(root=root, batch_size=4, num_classes=2)

        model = DummySegmentationBaseline(
            in_channels=dm.num_channels,
            num_classes=dm.num_classes,
            image_size=dm.image_size,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            optim_recipe=optim_cifar10_resnet18,
            log_plots=True,
            ema=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

        trainer = TUTrainer(
            accelerator="cpu",
            max_epochs=2,
            logger=None,
            enable_checkpointing=False,
        )

        root = Path(__file__).parent.absolute().parents[0] / "data"
        dm = DummySegmentationDataModule(root=root, batch_size=4, num_classes=2)

        model = DummySegmentationBaseline(
            in_channels=dm.num_channels,
            num_classes=dm.num_classes,
            image_size=dm.image_size,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            optim_recipe=optim_cifar10_resnet18,
            log_plots=True,
            swa=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_two_estimators_two_classes(self):
        trainer = TUTrainer(
            accelerator="cpu",
            max_epochs=2,
            logger=None,
            enable_checkpointing=False,
        )

        root = Path(__file__).parent.absolute().parents[0] / "data"
        dm = DummySegmentationDataModule(root=root, batch_size=4, num_classes=2)

        model = DummySegmentationBaseline(
            in_channels=dm.num_channels,
            num_classes=dm.num_classes,
            image_size=dm.image_size,
            loss=nn.CrossEntropyLoss(),
            baseline_type="ensemble",
            optim_recipe=optim_cifar10_resnet18,
            swa=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_segmentation_errors(self):
        with pytest.raises(ValueError, match="num_classes must be at least 2, got"):
            SegmentationRoutine(model=nn.Identity(), num_classes=1, loss=nn.CrossEntropyLoss())

        with pytest.raises(ValueError, match="metric_subsampling_rate must be in"):
            SegmentationRoutine(
                model=nn.Identity(),
                num_classes=2,
                loss=nn.CrossEntropyLoss(),
                metric_subsampling_rate=-1,
            )

        with pytest.raises(ValueError, match="num_calibration_bins must be at least 2, got"):
            SegmentationRoutine(
                model=nn.Identity(),
                num_classes=2,
                loss=nn.CrossEntropyLoss(),
                num_calibration_bins=0,
            )



import pytest
import torch
from torch.distributions import Normal

from torch_uncertainty.metrics import CategoricalNLL, DistributionNLL


class TestCategoricalNegativeLogLikelihood:
    """Testing the CategoricalNLL metric class."""

    def test_compute_zero(self) -> None:
        probs = torch.as_tensor([[1, 0.0], [0.0, 1.0]])
        targets = torch.as_tensor([0, 1])

        metric = CategoricalNLL()
        metric.update(probs, targets)
        res = metric.compute()
        assert res == 0

        metric = CategoricalNLL(reduction="none")
        metric.update(probs, targets)
        res_sum = metric.compute()
        assert torch.all(res_sum == torch.zeros(2))

        metric = CategoricalNLL(reduction="sum")
        metric.update(probs, targets)
        res_sum = metric.compute()
        assert torch.all(res_sum == torch.zeros(1))

    def test_bad_argument(self) -> None:
        with pytest.raises(ValueError):
            _ = CategoricalNLL(reduction="geometric_mean")


class TestDistributionNLL:
    """Testing the TestDistributionNLL metric class."""

    def test_compute_zero(self) -> None:
        metric = DistributionNLL(reduction="mean")
        means = torch.as_tensor([1, 10]).float()
        stds = torch.as_tensor([1, 2]).float()
        targets = torch.as_tensor([1, 10]).float()
        dist = Normal(means, stds)
        metric.update(dist, targets)
        res_mean = metric.compute()
        assert res_mean == torch.mean(torch.log(2 * torch.pi * (stds**2)) / 2)

        metric = DistributionNLL(reduction="sum")
        metric.update(dist, targets)
        res_sum = metric.compute()
        assert res_sum == torch.log(2 * torch.pi * (stds**2)).sum() / 2

        metric = DistributionNLL(reduction="none")
        metric.update(dist, targets)
        res_all = metric.compute()
        assert torch.all(res_all == torch.log(2 * torch.pi * (stds**2)) / 2)



from torch import Tensor, nn
from torch.nn.modules.batchnorm import _BatchNorm


class Backbone(nn.Module):
    def __init__(self, model: nn.Module, feat_names: list[str]) -> None:
        """Encoder backbone.

        Return the skip features of the :attr:`model` corresponding to the
        :attr:`feat_names`.

        Args:
            model (nn.Module): Base model.
            feat_names (list[str]): list of the feature names.
        """
        super().__init__()
        self.model = model
        self.feat_names = feat_names

    def forward(self, x: Tensor) -> list[Tensor]:
        """Encoder forward pass.

        Args:
            x (Tensor): Input tensor.

        Returns:
            list[Tensor]: list of the features.
        """
        feature = x
        features = []
        for key, layer in self.model._modules.items():
            feature = layer(feature)
            if key in self.feat_names:
                features.append(feature)
        return features


def set_bn_momentum(model: nn.Module, momentum: float) -> None:
    """Set the momentum of all batch normalization layers in the model.

    Args:
        model (nn.Module): Model.
        momentum (float): Momentum of the batch normalization layers.
    """
    for m in model.modules():
        if isinstance(m, _BatchNorm):
            m.momentum = momentum



import copy
from collections.abc import Mapping

import torch
from torch import Tensor, nn
from torch.utils.data import DataLoader

from .swa import SWA


class SWAG(SWA):
    swag_stats: dict[str, Tensor]
    prfx = "model.swag_stats."

    def __init__(
        self,
        model: nn.Module,
        cycle_start: int,
        cycle_length: int,
        scale: float = 1.0,
        diag_covariance: bool = False,
        max_num_models: int = 20,
        var_clamp: float = 1e-6,
        num_estimators: int = 16,
    ) -> None:
        """Stochastic Weight Averaging Gaussian (SWAG).

        Update the SWAG posterior every `cycle_length` epochs starting at
        `cycle_start`. Samples :attr:`num_estimators` models from the SWAG
        posterior after each update. Uses the SWAG posterior estimation only
        at test time. Otherwise, uses the base model for training.

        Call :meth:`update_wrapper` at the end of each epoch. It will update
        the SWAG posterior if the current epoch number minus :attr:`cycle_start`
        is a multiple of :attr:`cycle_length`. Call :meth:`bn_update` to update
        the batchnorm statistics of the current SWAG samples.

        Args:
            model (nn.Module): PyTorch model to be trained.
            cycle_start (int): Begininning of the first SWAG averaging cycle.
            cycle_length (int): Number of epochs between SWAG updates. The
                first update occurs at :attr:`cycle_start` + :attr:`cycle_length`.
            scale (float, optional): Scale of the Gaussian. Defaults to 1.0.
            diag_covariance (bool, optional): Whether to use a diagonal
                covariance. Defaults to False.
            max_num_models (int, optional): Maximum number of models to store.
                Defaults to 0.
            var_clamp (float, optional): Minimum variance. Defaults to 1e-30.
            num_estimators (int, optional): Number of posterior estimates to
                use. Defaults to 16.

        Reference:
            Maddox, W. J. et al. A simple baseline for bayesian uncertainty in
            deep learning. In NeurIPS 2019.

        Note:
            Originates from https://github.com/wjmaddox/swa_gaussian.
        """
        super().__init__(model, cycle_start, cycle_length)
        _swag_checks(scale, max_num_models, var_clamp)

        self.num_estimators = num_estimators
        self.scale = scale

        self.diag_covariance = diag_covariance
        self.max_num_models = max_num_models
        self.var_clamp = var_clamp

        self.initialize_stats()
        self.fit = False
        self.samples = []

    def eval_forward(self, x: Tensor) -> Tensor:
        """Forward pass of the SWAG model when in eval mode."""
        if not self.fit:
            return self.core_model.forward(x)
        return torch.cat([mod.to(device=x.device)(x) for mod in self.samples])

    def initialize_stats(self) -> None:
        """Initialize the SWAG dictionary of statistics.

        For each parameter, we create a mean, squared mean, and covariance
        square root. The covariance square root is only used when
        `diag_covariance` is False.
        """
        self.swag_stats = {}
        for name_p, param in self.core_model.named_parameters():
            mean, squared_mean = (
                torch.zeros_like(param, device="cpu"),
                torch.zeros_like(param, device="cpu"),
            )
            self.swag_stats[self.prfx + name_p + "_mean"] = mean
            self.swag_stats[self.prfx + name_p + "_sq_mean"] = squared_mean

            if not self.diag_covariance:
                covariance_sqrt = torch.zeros((0, param.numel()), device="cpu")
                self.swag_stats[self.prfx + name_p + "_covariance_sqrt"] = covariance_sqrt

    @torch.no_grad()
    def update_wrapper(self, epoch: int) -> None:
        """Update the SWAG posterior.

        The update is performed if the epoch is greater than the cycle start
        and the difference between the epoch and the cycle start is a multiple
        of the cycle length.

        Args:
            epoch (int): Current epoch.
        """
        if not (epoch > self.cycle_start and (epoch - self.cycle_start) % self.cycle_length == 0):
            return

        for name_p, param in self.core_model.named_parameters():
            mean = self.swag_stats[self.prfx + name_p + "_mean"]
            squared_mean = self.swag_stats[self.prfx + name_p + "_sq_mean"]
            new_param = param.data.detach().cpu()

            mean = mean * self.num_avgd_models / (self.num_avgd_models + 1) + new_param / (
                self.num_avgd_models + 1
            )
            squared_mean = squared_mean * self.num_avgd_models / (
                self.num_avgd_models + 1
            ) + new_param**2 / (self.num_avgd_models + 1)

            self.swag_stats[self.prfx + name_p + "_mean"] = mean
            self.swag_stats[self.prfx + name_p + "_sq_mean"] = squared_mean

            if not self.diag_covariance:
                covariance_sqrt = self.swag_stats[self.prfx + name_p + "_covariance_sqrt"]
                dev = (new_param - mean).view(-1, 1).t()
                covariance_sqrt = torch.cat((covariance_sqrt, dev), dim=0)
                if self.num_avgd_models + 1 > self.max_num_models:
                    covariance_sqrt = covariance_sqrt[1:, :]
                self.swag_stats[self.prfx + name_p + "_covariance_sqrt"] = covariance_sqrt

        self.num_avgd_models += 1

        self.samples = [
            self.sample(self.scale, self.diag_covariance) for _ in range(self.num_estimators)
        ]
        self.need_bn_update = True
        self.fit = True

    def bn_update(self, loader: DataLoader, device: torch.device) -> None:
        """Update the bachnorm statistics of the current SWAG samples.

        Args:
            loader (DataLoader): DataLoader to update the batchnorm statistics.
            device (torch.device): Device to perform the update.
        """
        if self.need_bn_update:
            for mod in self.samples:
                torch.optim.swa_utils.update_bn(loader, mod, device=device)
            self.need_bn_update = False

    def sample(
        self,
        scale: float,
        diag_covariance: bool | None = None,
        block: bool = False,
        seed: int | None = None,
    ) -> nn.Module:
        """Sample a model from the SWAG posterior.

        Args:
            scale (float): Rescale coefficient of the Gaussian.
            diag_covariance (bool, optional): Whether to use a diagonal
                covariance. Defaults to None.
            block (bool, optional): Whether to sample a block diagonal
                covariance. Defaults to False.
            seed (int, optional): Random seed. Defaults to None.

        Returns:
            nn.Module: Sampled model.
        """
        if seed is not None:
            torch.manual_seed(seed)

        if diag_covariance is None:
            diag_covariance = self.diag_covariance
        if not diag_covariance and self.diag_covariance:
            raise ValueError("Cannot sample full rank from diagonal covariance matrix.")

        if not block:
            return self._fullrank_sample(scale, diag_covariance)
        raise NotImplementedError("Raise an issue if you need this feature.")

    def _fullrank_sample(self, scale: float, diagonal_covariance: bool) -> nn.Module:
        new_sample = copy.deepcopy(self.core_model)

        for name_p, param in new_sample.named_parameters():
            mean = self.swag_stats[self.prfx + name_p + "_mean"]
            sq_mean = self.swag_stats[self.prfx + name_p + "_sq_mean"]

            if not diagonal_covariance:
                cov_mat_sqrt = self.swag_stats[self.prfx + name_p + "_covariance_sqrt"]

            var = torch.clamp(sq_mean - mean**2, self.var_clamp)
            var_sample = var.sqrt() * torch.randn_like(var, requires_grad=False)

            if not diagonal_covariance:
                cov_sample = cov_mat_sqrt.t() @ torch.randn((cov_mat_sqrt.size(0),))
                cov_sample /= (self.max_num_models - 1) ** 0.5
                var_sample += cov_sample.view_as(var_sample)

            sample = mean + scale**0.5 * var_sample
            param.data = sample.to(device="cpu", dtype=param.dtype)
        return new_sample

    def _save_to_state_dict(self, destination, prefix: str, keep_vars: bool):
        """Add the SWAG statistics to the destination dict."""
        super()._save_to_state_dict(destination, prefix, keep_vars)
        destination |= self.swag_stats

    def state_dict(self, *args, destination=None, prefix="", keep_vars=False) -> Mapping:
        """Add the SWAG statistics to the state dict."""
        return self.swag_stats | super().state_dict(
            *args, destination=destination, prefix=prefix, keep_vars=keep_vars
        )

    def _load_swag_stats(self, state_dict: Mapping):
        """Load the SWAG statistics from the state dict."""
        self.swag_stats = {k: v for k, v in state_dict.items() if k in self.swag_stats}
        for k in self.swag_stats:
            del state_dict[k]
        self.samples = [
            self.sample(self.scale, self.diag_covariance) for _ in range(self.num_estimators)
        ]
        self.need_bn_update = True
        self.fit = True

    def load_state_dict(self, state_dict: Mapping, strict: bool = True, assign: bool = False):
        self._load_swag_stats(state_dict)
        return super().load_state_dict(state_dict, strict, assign)

    def compute_logdet(self, block=False):
        raise NotImplementedError("Raise an issue if you need this feature.")

    def compute_logprob(self, vec=None, block=False, diag=False):
        raise NotImplementedError("Raise an issue if you need this feature.")


def _swag_checks(scale: float, max_num_models: int, var_clamp: float) -> None:
    if scale < 0:
        raise ValueError(f"`scale` must be non-negative. Got {scale}.")
    if max_num_models < 0:
        raise ValueError(f"`max_num_models` must be non-negative. Got {max_num_models}.")
    if var_clamp < 0:
        raise ValueError(f"`var_clamp` must be non-negative. Got {var_clamp}.")



# ruff: noqa: F401
from .baseline import (
    DummyClassificationBaseline,
    DummyPixelRegressionBaseline,
    DummyRegressionBaseline,
    DummySegmentationBaseline,
)
from .datamodule import (
    DummyClassificationDataModule,
    DummyPixelRegressionDataModule,
    DummyRegressionDataModule,
    DummySegmentationDataModule,
)
from .dataset import (
    DummPixelRegressionDataset,
    DummyClassificationDataset,
    DummyRegressionDataset,
    DummySegmentationDataset,
)
from .model import dummy_model
from .transform import DummyTransform



from typing import Literal

import matplotlib.cm as cm
import torch
from einops import rearrange
from lightning.pytorch import LightningModule
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch import Tensor, nn
from torch.distributions import (
    Categorical,
    Distribution,
    Independent,
    MixtureSameFamily,
)
from torch.optim import Optimizer
from torchmetrics import MeanSquaredError, MetricCollection
from torchvision.transforms.v2 import functional as F
from torchvision.utils import make_grid

from torch_uncertainty.metrics import (
    DistributionNLL,
    Log10,
    MeanAbsoluteErrorInverse,
    MeanGTRelativeAbsoluteError,
    MeanGTRelativeSquaredError,
    MeanSquaredErrorInverse,
    MeanSquaredLogError,
    SILog,
    ThresholdAccuracy,
)
from torch_uncertainty.models import (
    EPOCH_UPDATE_MODEL,
    STEP_UPDATE_MODEL,
)
from torch_uncertainty.utils.distributions import (
    get_dist_class,
    get_dist_estimate,
)


class PixelRegressionRoutine(LightningModule):
    inv_norm_params = {
        "mean": [-0.485 / 0.229, -0.456 / 0.224, -0.406 / 0.255],
        "std": [1 / 0.229, 1 / 0.224, 1 / 0.255],
    }

    def __init__(
        self,
        model: nn.Module,
        output_dim: int,
        loss: nn.Module,
        dist_family: str | None = None,
        dist_estimate: str = "mean",
        is_ensemble: bool = False,
        format_batch_fn: nn.Module | None = None,
        optim_recipe: dict | Optimizer | None = None,
        eval_shift: bool = False,
        num_image_plot: int = 4,
        log_plots: bool = False,
    ) -> None:
        r"""Routine for training & testing on **pixel regression** tasks.

        Args:
            model (nn.Module): Model to train.
            output_dim (int): Number of outputs of the model.
            loss (nn.Module): Loss function to optimize the :attr:`model`.
            dist_family (str, optional): The distribution family to use for
                probabilistic pixel regression. If ``None`` then point-wise regression.
                Defaults to ``None``.
            dist_estimate (str, optional): The estimate to use when computing the
                point-wise metrics. Defaults to ``"mean"``.
            is_ensemble (bool, optional): Whether the model is an ensemble.
                Defaults to ``False``.
            optim_recipe (dict or Optimizer, optional): The optimizer and
                optionally the scheduler to use. Defaults to ``None``.
            eval_shift (bool, optional): Indicates whether to evaluate the Distribution
                shift performance. Defaults to ``False``.
            format_batch_fn (nn.Module, optional): The function to format the
                batch. Defaults to ``None``.
            num_image_plot (int, optional): Number of images to plot. Defaults to ``4``.
            log_plots (bool, optional): Indicates whether to log plots from
                metrics. Defaults to ``False``.
        """
        super().__init__()
        _depth_routine_checks(output_dim, num_image_plot, log_plots)
        if eval_shift:
            raise NotImplementedError(
                "Distribution shift evaluation not implemented yet. Raise an issue if needed."
            )

        self.model = model
        self.output_dim = output_dim
        self.one_dim_depth = output_dim == 1
        self.dist_family = dist_family
        self.dist_estimate = dist_estimate
        self.probabilistic = dist_family is not None
        self.loss = loss
        self.num_image_plot = num_image_plot
        self.is_ensemble = is_ensemble
        self.log_plots = log_plots

        self.needs_epoch_update = isinstance(model, EPOCH_UPDATE_MODEL)
        self.needs_step_update = isinstance(model, STEP_UPDATE_MODEL)

        if format_batch_fn is None:
            format_batch_fn = nn.Identity()

        self.optim_recipe = optim_recipe
        self.format_batch_fn = format_batch_fn
        self._init_metrics()

    def _init_metrics(self) -> None:
        """Initialize the metrics depending on the exact task."""
        depth_metrics = MetricCollection(
            {
                "reg/SILog": SILog(),
                "reg/log10": Log10(),
                "reg/ARE": MeanGTRelativeAbsoluteError(),
                "reg/RSRE": MeanGTRelativeSquaredError(squared=False),
                "reg/RMSE": MeanSquaredError(squared=False),
                "reg/RMSELog": MeanSquaredLogError(squared=False),
                "reg/iMAE": MeanAbsoluteErrorInverse(),
                "reg/iRMSE": MeanSquaredErrorInverse(squared=False),
                "reg/d1": ThresholdAccuracy(power=1),
                "reg/d2": ThresholdAccuracy(power=2),
                "reg/d3": ThresholdAccuracy(power=3),
            },
            compute_groups=False,
        )

        self.val_metrics = depth_metrics.clone(prefix="val/")
        self.test_metrics = depth_metrics.clone(prefix="test/")

        if self.probabilistic:
            depth_prob_metrics = MetricCollection({"reg/NLL": DistributionNLL(reduction="mean")})
            self.val_prob_metrics = depth_prob_metrics.clone(prefix="val/")
            self.test_prob_metrics = depth_prob_metrics.clone(prefix="test/")

    def configure_optimizers(self) -> Optimizer | dict:
        return self.optim_recipe

    def on_train_start(self) -> None:
        """Put the hyperparameters in tensorboard."""
        if self.logger is not None:  # coverage: ignore
            self.logger.log_hyperparams(
                self.hparams,
            )

    def on_validation_start(self) -> None:
        """Prepare the validation step.

        Update the model's wrapper and the batchnorms if needed.
        """
        if self.needs_epoch_update and not self.trainer.sanity_checking:
            self.model.update_wrapper(self.current_epoch)
            if hasattr(self.model, "need_bn_update"):
                self.model.bn_update(self.trainer.train_dataloader, device=self.device)

    def on_test_start(self) -> None:
        """Prepare the test step.

        Update the batchnorms if needed.
        """
        if hasattr(self.model, "need_bn_update"):
            self.model.bn_update(self.trainer.train_dataloader, device=self.device)

    def forward(self, inputs: Tensor) -> Tensor | Distribution:
        """Forward pass of the routine.

        The forward pass automatically squeezes the output if the regression
        is one-dimensional and if the routine contains a single model.

        Args:
            inputs (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor.
        """
        pred = self.model(inputs)
        if self.probabilistic:
            if not self.is_ensemble:
                pred = {k: v.squeeze(-1) for k, v in pred.items()}
        else:
            if not self.is_ensemble:
                pred = pred.squeeze(-1)
        return pred

    def training_step(self, batch: tuple[Tensor, Tensor]) -> STEP_OUTPUT:
        """Perform a single training step based on the input tensors.

        Args:
            batch (tuple[Tensor, Tensor]): the training data and their corresponding targets

        Returns:
            Tensor: the loss corresponding to this training step.
        """
        inputs, target = self.format_batch_fn(batch)
        if self.one_dim_depth:
            target = target.unsqueeze(1)

        out = self.model(inputs)
        out_shape = out[next(iter(out))].shape[-2:] if self.probabilistic else out.shape[-2:]
        target = F.resize(target, out_shape, interpolation=F.InterpolationMode.NEAREST)
        target = rearrange(target, "b c h w -> b h w c")
        padding_mask = torch.isnan(target).any(dim=-1)
        if self.probabilistic:
            dist_params = {k: rearrange(v, "b c h w -> b h w c") for k, v in out.items()}
            # Adding the Independent wrapper to the distribution to compute correctly the
            # log-likelihood given a target. Here the last dimension is the event dimension.
            # When computing the log-likelihood, the values are summed over the event dimension.
            dists = Independent(get_dist_class(self.dist_family)(**dist_params), 1)
            loss = self.loss(dists, target, padding_mask)
        else:
            out = rearrange(out, "b c h w -> b h w c")
            loss = self.loss(out[padding_mask], target[padding_mask])

        if self.needs_step_update:
            self.model.update_wrapper(self.current_epoch)
        self.log("train_loss", loss, prog_bar=True, logger=True)
        return loss

    def evaluation_forward(self, inputs: Tensor) -> tuple[Tensor, Distribution | None]:
        """Get the prediction and handle predicted eventual distribution parameters.

        Args:
            inputs (Tensor): the input data.

        Returns:
            tuple[Tensor, Distribution | None]: the prediction as a Tensor and a distribution.
        """
        batch_size = inputs.size(0)
        preds = self.model(inputs)

        if self.probabilistic:
            dist_params = {
                k: rearrange(v, "(m b) c h w -> b h w m c", b=batch_size) for k, v in preds.items()
            }
            # Adding the Independent wrapper to the distribution to create a MixtureSameFamily.
            # As required by the torch.distributions API, the last dimension is the event dimension.
            comp = Independent(get_dist_class(self.dist_family)(**dist_params), 1)
            mix = Categorical(torch.ones(comp.batch_shape, device=self.device))
            mixture = MixtureSameFamily(mix, comp)
            preds = get_dist_estimate(comp, self.dist_estimate).mean(-2)
            return preds, mixture

        preds = rearrange(preds, "(m b) c h w -> b m h w c", b=batch_size)
        return preds.mean(dim=1), None

    def validation_step(self, batch: tuple[Tensor, Tensor], batch_idx: int) -> None:
        """Perform a single validation step based on the input tensors.

        Compute the prediction of the model and the value of the metrics on the validation batch.

        Args:
            batch (tuple[Tensor, Tensor]): the validation images and their corresponding targets.
            batch_idx (int): the id of the batch. Optionally plot images and the predictions with
                the first batch.
        """
        inputs, targets = batch
        if self.one_dim_depth:
            targets = targets.unsqueeze(1)
        targets = rearrange(targets, "b c h w -> b h w c")
        preds, dist = self.evaluation_forward(inputs)

        if batch_idx == 0 and self.log_plots:
            self._plot_depth(
                inputs[: self.num_image_plot, ...],
                preds[: self.num_image_plot, ...],
                targets[: self.num_image_plot, ...],
                stage="val",
            )

        padding_mask = torch.isnan(targets).any(dim=-1)
        self.val_metrics.update(preds[padding_mask], targets[padding_mask])
        if isinstance(dist, Distribution):
            self.val_prob_metrics.update(dist, targets, padding_mask)

    def test_step(
        self,
        batch: tuple[Tensor, Tensor],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Perform a single test step based on the input tensors.

        Compute the prediction of the model and the value of the metrics on the test batch. Also
        handle OOD and distribution-shifted images.

        Args:
            batch (tuple[Tensor, Tensor]): the test data and their corresponding targets.
            batch_idx (int): the number of the current batch (unused).
            dataloader_idx (int): 0 if in-distribution, 1 if out-of-distribution.
        """
        if dataloader_idx != 0:
            raise NotImplementedError(
                "Depth OOD detection not implemented yet. Raise an issue if needed."
            )
        inputs, targets = batch
        if self.one_dim_depth:
            targets = targets.unsqueeze(1)
        targets = rearrange(targets, "b c h w -> b h w c")
        preds, dist = self.evaluation_forward(inputs)

        if batch_idx == 0 and self.log_plots:
            num_images = (
                self.num_image_plot if self.num_image_plot < inputs.size(0) else inputs.size(0)
            )
            self._plot_depth(
                inputs[:num_images, ...],
                preds[:num_images, ...],
                targets[:num_images, ...],
                stage="test",
            )

        padding_mask = torch.isnan(targets).any(dim=-1)
        self.test_metrics.update(preds[padding_mask], targets[padding_mask])
        if isinstance(dist, Distribution):
            self.test_prob_metrics.update(dist, targets, padding_mask)

    def on_validation_epoch_end(self) -> None:
        """Compute and log the values of the collected metrics in `validation_step`."""
        res_dict = self.val_metrics.compute()
        self.log_dict(res_dict, logger=True, sync_dist=True)
        self.log(
            "RMSE",
            res_dict["val/reg/RMSE"],
            prog_bar=True,
            logger=False,
            sync_dist=True,
        )
        self.val_metrics.reset()
        if self.probabilistic:
            self.log_dict(
                self.val_prob_metrics.compute(),
                sync_dist=True,
            )
            self.val_prob_metrics.reset()

    def on_test_epoch_end(self) -> None:
        """Compute and log the values of the collected metrics in `test_step`."""
        self.log_dict(
            self.test_metrics.compute(),
            sync_dist=True,
        )
        self.test_metrics.reset()
        if self.probabilistic:
            self.log_dict(
                self.test_prob_metrics.compute(),
                sync_dist=True,
            )
            self.test_prob_metrics.reset()

    def _plot_depth(
        self,
        inputs: Tensor,
        preds: Tensor,
        target: Tensor,
        stage: Literal["val", "test"],
    ) -> None:
        if (
            self.logger is not None
            and isinstance(self.logger, TensorBoardLogger)
            and self.one_dim_depth
        ):
            all_imgs = []
            for i in range(inputs.size(0)):
                img = F.normalize(inputs[i, ...].cpu(), **self.inv_norm_params)
                pred = colorize(preds[i, 0, ...].cpu(), vmin=0, vmax=self.model.max_depth)
                tgt = colorize(target[i, 0, ...].cpu(), vmin=0, vmax=self.model.max_depth)
                all_imgs.extend([img, pred, tgt])

            self.logger.experiment.add_image(
                f"{stage}/samples",
                make_grid(torch.stack(all_imgs, dim=0), nrow=3),
                self.current_epoch,
            )


def colorize(
    value: Tensor,
    vmin: float | None = None,
    vmax: float | None = None,
    cmap: str = "magma",
):
    """Colorize a tensor of depth values.

    Args:
        value (Tensor): The tensor of depth values.
        vmin (float, optional): The minimum depth value. Defaults to None.
        vmax (float, optional): The maximum depth value. Defaults to None.
        cmap (str, optional): The colormap to use. Defaults to 'magma'.
    """
    vmin = value.min().item() if vmin is None else vmin
    vmax = value.max().item() if vmax is None else vmax
    if vmin == vmax:
        return torch.zeros_like(value)
    value = (value - vmin) / (vmax - vmin)
    cmapper = cm.get_cmap(cmap)
    value = cmapper(value.numpy(), bytes=True)
    img = value[:, :, :3]
    return torch.as_tensor(img).permute(2, 0, 1).float() / 255.0


def _depth_routine_checks(output_dim: int, num_image_plot: int, log_plots: bool) -> None:
    """Check the domains of the routine's parameters.

    Args:
        output_dim (int): the dimension of the output of the regression task.
        num_image_plot (int): the number of images to plot at evaluation time.
        log_plots (bool): whether to plot images and predictions during evaluation.
    """
    if output_dim < 1:
        raise ValueError(f"output_dim must be positive, got {output_dim}.")
    if num_image_plot < 1 and log_plots:
        raise ValueError(f"num_image_plot must be positive, got {num_image_plot}.")



import pytest
import torch

from torch_uncertainty.metrics import BrierScore


@pytest.fixture()
def vec2d_max() -> torch.Tensor:
    vec = torch.as_tensor([0.5, 0.5])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec2d_max_target() -> torch.Tensor:
    vec = torch.as_tensor([0, 1])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec2d_max_target1d() -> torch.Tensor:
    return torch.as_tensor([1])


@pytest.fixture()
def vec2d_min() -> torch.Tensor:
    vec = torch.as_tensor([0.0, 1.0])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec2d_min_target() -> torch.Tensor:
    vec = torch.as_tensor([0, 1])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec2d_5classes() -> torch.Tensor:
    return torch.as_tensor([[0.2, 0.6, 0.1, 0.05, 0.05], [0.05, 0.25, 0.1, 0.3, 0.3]])


@pytest.fixture()
def vec2d_5classes_target() -> torch.Tensor:
    return torch.as_tensor([[0, 0, 0, 1, 0], [0, 0, 0, 0, 1]])


@pytest.fixture()
def vec2d_5classes_target1d() -> torch.Tensor:
    return torch.as_tensor([3, 4])


@pytest.fixture()
def vec3d() -> torch.Tensor:
    """Return a torch tensor with a mean BrierScore of 0 and a BrierScore of
    the mean of 0.5 to test the `ensemble` parameter of `BrierScore`.
    """
    vec = torch.as_tensor([[0.0, 1.0], [1.0, 0.0]])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec3d_target() -> torch.Tensor:
    vec = torch.as_tensor([0, 1])
    return vec.unsqueeze(0)


@pytest.fixture()
def vec3d_target1d() -> torch.Tensor:
    vec = torch.as_tensor([1])
    return vec.unsqueeze(0)


class TestBrierScore:
    """Testing the BrierScore metric class."""

    def test_compute(self, vec2d_min: torch.Tensor, vec2d_min_target: torch.Tensor):
        metric = BrierScore(num_classes=2)
        metric.update(vec2d_min, vec2d_min_target)
        assert metric.compute() == 0

        metric = BrierScore(num_classes=2, top_class=True)
        metric.update(vec2d_min, vec2d_min_target)
        assert metric.compute() == 0

    def test_compute_max(self, vec2d_max: torch.Tensor, vec2d_max_target: torch.Tensor):
        metric = BrierScore(num_classes=2, reduction="sum")
        metric.update(vec2d_max, vec2d_max_target)
        assert metric.compute() == 0.5

    def test_compute_max_target1d(self, vec2d_max: torch.Tensor, vec2d_max_target1d: torch.Tensor):
        metric = BrierScore(num_classes=2, reduction="sum")
        metric.update(vec2d_max, vec2d_max_target1d)
        assert metric.compute() == 0.5

    def test_compute_5classes(
        self,
        vec2d_5classes: torch.Tensor,
        vec2d_5classes_target: torch.Tensor,
        vec2d_5classes_target1d: torch.Tensor,
    ):
        metric = BrierScore(num_classes=5, reduction="sum")
        metric.update(vec2d_5classes, vec2d_5classes_target)
        metric.update(vec2d_5classes, vec2d_5classes_target1d)
        assert (
            metric.compute() / 2
            == 0.2**2 + 0.6**2 + 0.1**2 * 2 + 0.95**2 + 0.05**2 * 2 + 0.25**2 + 0.3**2 + 0.7**2
        )

        metric = BrierScore(num_classes=5, top_class=True, reduction="sum")
        metric.update(vec2d_5classes, vec2d_5classes_target)
        assert metric.compute() == pytest.approx(0.6**2 + 0.3**2)

    def test_multiple_compute_sum(
        self,
        vec2d_min: torch.Tensor,
        vec2d_max: torch.Tensor,
        vec2d_min_target: torch.Tensor,
        vec2d_max_target: torch.Tensor,
    ):
        metric = BrierScore(num_classes=2, reduction="sum")
        metric.update(vec2d_min, vec2d_min_target)
        metric.update(vec2d_max, vec2d_max_target)
        assert metric.compute() == 0.5

    def test_multiple_compute_mean(
        self,
        vec2d_min: torch.Tensor,
        vec2d_max: torch.Tensor,
        vec2d_min_target: torch.Tensor,
        vec2d_max_target: torch.Tensor,
    ):
        metric = BrierScore(num_classes=2, reduction="mean")
        metric.update(vec2d_min, vec2d_min_target)
        metric.update(vec2d_max, vec2d_max_target)
        assert metric.compute() == 0.5 / 2

    def test_multiple_compute_none(
        self,
        vec2d_min: torch.Tensor,
        vec2d_max: torch.Tensor,
        vec2d_min_target: torch.Tensor,
        vec2d_max_target: torch.Tensor,
    ):
        metric = BrierScore(num_classes=2, reduction=None)
        metric.update(vec2d_min, vec2d_min_target)
        metric.update(vec2d_max, vec2d_max_target)
        assert all(metric.compute() == torch.as_tensor([0, 0.5]))

    def test_compute_3d_mean(self, vec3d: torch.Tensor, vec3d_target: torch.Tensor):
        """Test that the metric returns the mean of the BrierScore over
        the estimators.
        """
        metric = BrierScore(num_classes=2, reduction="mean")
        metric.update(vec3d, vec3d_target)
        assert metric.compute() == 1

    def test_compute_3d_sum(self, vec3d: torch.Tensor, vec3d_target: torch.Tensor):
        metric = BrierScore(num_classes=2, reduction="sum")
        metric.update(vec3d, vec3d_target)
        assert metric.compute() == 1

    def test_compute_3d_sum_target1d(self, vec3d: torch.Tensor, vec3d_target1d: torch.Tensor):
        metric = BrierScore(num_classes=2, reduction="sum")
        metric.update(vec3d, vec3d_target1d)
        assert metric.compute() == 1

    def test_compute_3d_to_2d(self, vec3d: torch.Tensor, vec3d_target: torch.Tensor):
        metric = BrierScore(num_classes=2, reduction="mean")
        vec3d = vec3d.mean(1)
        metric.update(vec3d, vec3d_target)
        assert metric.compute() == 0.5

    def test_bad_input(self) -> None:
        with pytest.raises(ValueError):
            metric = BrierScore(num_classes=2, reduction="none")
            metric.update(torch.ones(2, 2, 2, 2), torch.ones(2, 2, 2, 2))

    def test_bad_argument(self):
        with pytest.raises(ValueError, match="Expected argument `reduction` to be one of"):
            _ = BrierScore(num_classes=2, reduction="geometric_mean")



from pathlib import Path

from torch_uncertainty import cli_main, init_args
from torch_uncertainty.baselines import DeepEnsemblesBaseline
from torch_uncertainty.datamodules import CIFAR100DataModule

if __name__ == "__main__":
    args = init_args(DeepEnsemblesBaseline, CIFAR100DataModule)
    if args.root == "./data/":
        root = Path(__file__).parent.absolute().parents[2]
    else:
        root = Path(args.root)

    net_name = f"de-{args.backbone}-cifar100"

    # datamodule
    args.root = str(root / "data")
    dm = CIFAR100DataModule(**vars(args))

    # model
    args.task = "classification"
    model = DeepEnsemblesBaseline(
        **vars(args),
        num_classes=dm.num_classes,
        in_channels=dm.num_channels,
    )

    args.test = -1

    cli_main(model, dm, root, net_name, args)



from collections.abc import Callable
from pathlib import Path
from typing import Any

import numpy as np
import torch
from torchvision.datasets import CIFAR10
from torchvision.datasets.utils import check_integrity, download_url


class CIFAR10H(CIFAR10):
    """`CIFAR-10H <https://github.com/jcpeterson/cifar-10h>`_ Dataset.

    Args:
        root (string): Root directory of dataset where file
            ``cifar-10h-probs.npy`` exists or will be saved to if download
            is set to True.
        train (bool, optional): For API consistency, not used.
        transform (callable, optional): A function/transform that takes in
            a PIL image and returns a transformed version. E.g,
            ``transforms.RandomCrop``. Defaults to None.
        target_transform (callable, optional): A function/transform that
            takes in the target and transforms it. Defaults to None.
        download (bool, optional): If True, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again. Defaults to False.
    """

    h_test_list = ["cifar-10h-probs.npy", "7b41f73eee90fdefc73bfc820ab29ba8"]
    h_url = "https://github.com/jcpeterson/cifar-10h/raw/master/data/cifar10h-probs.npy"

    def __init__(
        self,
        root: str | Path,
        train: bool | None = None,
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        download: bool = False,
    ) -> None:
        if train:
            raise ValueError("CIFAR10H does not support training data.")
        print("WARNING: CIFAR10H cannot be used within Classification routines for now.")
        super().__init__(
            Path(root),
            train=False,
            transform=transform,
            target_transform=target_transform,
            download=download,
        )

        if download:
            self.download_h()

        if not self._check_specific_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        self.targets = list(torch.as_tensor(np.load(self.root / self.h_test_list[0])))

    def _check_specific_integrity(self) -> bool:
        filename, md5 = self.h_test_list
        fpath = self.root / filename
        return check_integrity(fpath, md5)

    def download_h(self) -> None:
        download_url(
            self.h_url,
            self.root,
            filename=self.h_test_list[0],
            md5=self.h_test_list[1],
        )



# ruff: noqa: F401
from .batch_ensemble import BatchEnsemble, batch_ensemble
from .checkpoint_ensemble import (
    CheckpointEnsemble,
)
from .deep_ensembles import deep_ensembles
from .ema import EMA
from .mc_dropout import MCDropout, mc_dropout
from .stochastic import StochasticModel
from .swa import SWA
from .swag import SWAG

STEP_UPDATE_MODEL = (EMA,)
EPOCH_UPDATE_MODEL = (SWA, SWAG, CheckpointEnsemble)



import pytest
from torchvision.datasets import CIFAR10

from tests._dummies.dataset import DummyClassificationDataset
from torch_uncertainty.datamodules import CIFAR10DataModule
from torch_uncertainty.transforms import Cutout


class TestCIFAR10DataModule:
    """Testing the CIFAR10DataModule datamodule class."""

    def test_cifar10_main(self):
        dm = CIFAR10DataModule(root="./data/", batch_size=128, cutout=16)

        assert dm.dataset == CIFAR10
        assert isinstance(dm.train_transform.transforms[2], Cutout)

        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.shift_dataset = DummyClassificationDataset

        dm.prepare_data()
        dm.setup()

        with pytest.raises(ValueError):
            dm.setup("xxx")

        # test abstract methods
        dm.get_train_set()
        dm.get_val_set()
        dm.get_test_set()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.eval_ood = True
        dm.eval_shift = True
        dm.prepare_data()
        dm.setup("test")
        dm.test_dataloader()

        dm = CIFAR10DataModule(
            root="./data/",
            batch_size=128,
            cutout=16,
            test_alt="h",
            basic_augment=False,
        )
        dm.dataset = DummyClassificationDataset
        dm.setup("test")

        dm = CIFAR10DataModule(
            root="./data/",
            batch_size=128,
            cutout=16,
            num_dataloaders=2,
            val_split=0.1,
        )
        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.setup()
        dm.train_dataloader()

        with pytest.raises(ValueError):
            dm = CIFAR10DataModule(
                root="./data/",
                batch_size=128,
                cutout=8,
                num_dataloaders=2,
                val_split=0.1,
                auto_augment="rand-m9-n2-mstd0.5",
            )

        with pytest.raises(ValueError, match="CIFAR-H can only be used in testing."):
            dm = CIFAR10DataModule(
                root="./data/",
                batch_size=128,
                test_alt="h",
            )
            dm.setup("fit")

        with pytest.raises(ValueError, match="Test set "):
            dm = CIFAR10DataModule(
                root="./data/",
                batch_size=128,
                test_alt="x",
            )

        dm = CIFAR10DataModule(
            root="./data/",
            batch_size=128,
            cutout=None,
            num_dataloaders=2,
            val_split=0.1,
            auto_augment="rand-m9-n2-mstd0.5",
        )

    def test_cifar10_cv(self):
        dm = CIFAR10DataModule(root="./data/", batch_size=128)
        dm.dataset = lambda root, train, download, transform: DummyClassificationDataset(
            root,
            train=train,
            download=download,
            transform=transform,
            num_images=20,
        )
        dm.make_cross_val_splits(2, 1)

        dm = CIFAR10DataModule(root="./data/", batch_size=128, val_split=0.1)
        dm.dataset = lambda root, train, download, transform: DummyClassificationDataset(
            root,
            train=train,
            download=download,
            transform=transform,
            num_images=20,
        )
        dm.make_cross_val_splits(2, 1)



from pathlib import Path


def get_version(root: str | Path, version: int, checkpoint: int | None = None) -> tuple[Path, Path]:
    """Find the path to the checkpoint corresponding to the version.

    Args:
        root (Union[str, Path]): The root of the dataset containing the
            checkpoints.
        version (int): The version of the checkpoint.
        checkpoint (int, optional): The number of the checkpoint. Defaults
            to None.

    Raises:
        FileNotFoundError: if the checkpoint cannot be found.

    Returns:
        Tuple[Path, Path]: The path to the checkpoints and to its parameters.
    """
    root = Path(root)

    if (root / f"version_{version}").is_dir():
        version_folder = root / f"version_{version}"
        ckpt_folder = version_folder / "checkpoints"
        if checkpoint is None:
            ckpts = list(ckpt_folder.glob("*.ckpt"))
        else:
            ckpts = list(ckpt_folder.glob(f"epoch={checkpoint}-*.ckpt"))
    else:
        raise FileNotFoundError(f"The directory {root}/version_{version} does not exist.")

    file = ckpts[0]
    return (file.resolve(), (version_folder / "hparams.yaml").resolve())



import pytest

from tests._dummies.dataset import DummPixelRegressionDataset
from torch_uncertainty.datamodules.depth import (
    KITTIDataModule,
    MUADDataModule,
    NYUv2DataModule,
)
from torch_uncertainty.datasets import MUAD, KITTIDepth, NYUv2


class TestMUADDataModule:
    """Testing the MUADDataModule datamodule."""

    def test_muad_main(self):
        dm = MUADDataModule(root="./data/", min_depth=0, max_depth=100, batch_size=128)

        assert dm.dataset == MUAD

        dm.dataset = DummPixelRegressionDataset

        dm.prepare_data()
        dm.setup()

        with pytest.raises(ValueError):
            dm.setup("xxx")

        # test abstract methods
        dm.get_train_set()
        dm.get_val_set()
        dm.get_test_set()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.val_split = 0.1
        dm.prepare_data()
        dm.setup()
        dm.train_dataloader()
        dm.val_dataloader()


class TestNYUDataModule:
    """Testing the MUADDataModule datamodule."""

    def test_nyu_main(self):
        dm = NYUv2DataModule(root="./data/", max_depth=100, batch_size=128)

        assert dm.dataset == NYUv2

        dm.dataset = DummPixelRegressionDataset

        dm.prepare_data()
        dm.setup()

        with pytest.raises(ValueError):
            dm.setup("xxx")

        # test abstract methods
        dm.get_train_set()
        dm.get_val_set()
        dm.get_test_set()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.val_split = 0.1
        dm.prepare_data()
        dm.setup()
        dm.train_dataloader()
        dm.val_dataloader()

    def test_kitti_main(self):
        dm = KITTIDataModule(root="./data/", max_depth=100, batch_size=128)
        assert dm.dataset == KITTIDepth



import pytest
import torch
from einops import repeat

from torch_uncertainty.layers.functional.packed import (
    packed_in_projection_packed,
    packed_multi_head_attention_forward,
)
from torch_uncertainty.layers.packed import (
    PackedConv1d,
    PackedConv2d,
    PackedConv3d,
    PackedLayerNorm,
    PackedLinear,
    PackedMultiheadAttention,
    PackedTransformerDecoderLayer,
    PackedTransformerEncoderLayer,
)


@pytest.fixture()
def feat_input() -> torch.Tensor:
    return torch.rand((6, 1))  # (Cin, Lin)


@pytest.fixture()
def feat_input_one_rearrange() -> torch.Tensor:
    return torch.rand((1 * 3, 5))


@pytest.fixture()
def feat_multi_dim() -> torch.Tensor:
    return torch.rand((1, 2, 3, 4, 6))


@pytest.fixture()
def feat_input_16_features() -> torch.Tensor:
    return torch.rand((3, 16))


@pytest.fixture()
def seq_input() -> torch.Tensor:
    return torch.rand((5, 6, 3))


@pytest.fixture()
def img_input() -> torch.Tensor:
    return torch.rand((5, 6, 3, 3))


@pytest.fixture()
def voxels_input() -> torch.Tensor:
    return torch.rand((5, 6, 3, 3, 3))


@pytest.fixture()
def unbatched_qkv() -> torch.Tensor:
    return torch.rand((3, 6))


@pytest.fixture()
def unbatched_q_kv() -> tuple[torch.Tensor, torch.Tensor]:
    return torch.rand((3, 6)), torch.rand((4, 2))


@pytest.fixture()
def unbatched_q_k_v() -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    return torch.rand((3, 6)), torch.rand((4, 2)), torch.rand((4, 4))


@pytest.fixture()
def batched_qkv() -> torch.Tensor:
    return torch.rand((2, 3, 6))


@pytest.fixture()
def extended_batched_qkv() -> torch.Tensor:
    expansion = 2
    return torch.rand((2, 3, 6 * expansion))


@pytest.fixture()
def batched_q_kv() -> tuple[torch.Tensor, torch.Tensor]:
    return (
        torch.rand((2, 3, 6)),
        torch.rand((2, 4, 2)),
    )


@pytest.fixture()
def batched_q_k_v() -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    return (
        torch.rand((2, 3, 6)),
        torch.rand((2, 4, 2)),
        torch.rand((2, 4, 4)),
    )


@pytest.fixture()
def extended_batched_q_k_v() -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    expansion = 2
    return (
        torch.rand((2, 3, 6 * expansion)),
        torch.rand((2, 4, 2 * expansion)),
        torch.rand((2, 4, 4 * expansion)),
    )


@pytest.fixture()
def unbatched_tgt_memory() -> tuple[torch.Tensor, torch.Tensor]:
    return torch.rand((3, 6)), torch.rand((4, 6))


@pytest.fixture()
def batched_tgt_memory() -> tuple[torch.Tensor, torch.Tensor]:
    return (
        torch.rand((2, 3, 6)),
        torch.rand((2, 4, 6)),
    )


@pytest.fixture()
def extended_batched_tgt_memory() -> tuple[torch.Tensor, torch.Tensor]:
    expansion = 2
    return (
        torch.rand((2, 3, 6 * expansion)),
        torch.rand((2, 4, 6 * expansion)),
    )


class TestPackedLinear:
    """Testing the PackedLinear layer class."""

    # Legacy tests
    def test_linear_one_estimator_no_rearrange(self, feat_input: torch.Tensor):
        layer = PackedLinear(6, 2, alpha=1, num_estimators=1, rearrange=False, bias=False)
        out = layer(feat_input)
        assert out.shape == torch.Size([2, 1])

    def test_linear_two_estimators_no_rearrange(self, feat_input: torch.Tensor):
        layer = PackedLinear(6, 2, alpha=1, num_estimators=2, rearrange=False)
        out = layer(feat_input)
        assert out.shape == torch.Size([2, 1])

    def test_linear_one_estimator_rearrange(self, feat_input_one_rearrange: torch.Tensor):
        layer = PackedLinear(5, 2, alpha=1, num_estimators=1, rearrange=True)
        out = layer(feat_input_one_rearrange)
        assert out.shape == torch.Size([3, 2])

    def test_linear_two_estimator_rearrange_not_divisible(self):
        feat = torch.rand((2 * 3, 3))
        layer = PackedLinear(5, 1, alpha=1, num_estimators=2, rearrange=True)
        out = layer(feat)
        assert out.shape == torch.Size([6, 1])

    # Full implementation tests
    def test_linear_full_implementation(
        self, feat_input_16_features: torch.Tensor, feat_multi_dim: torch.Tensor
    ):
        layer = PackedLinear(16, 4, alpha=1, num_estimators=1, implementation="full", bias=False)
        out = layer(feat_input_16_features)
        assert out.shape == torch.Size([3, 4])
        layer = PackedLinear(16, 4, alpha=1, num_estimators=2, implementation="full")
        out = layer(feat_input_16_features)
        assert out.shape == torch.Size([3, 4])
        layer = PackedLinear(6, 2, alpha=1, num_estimators=1, implementation="full")
        out = layer(feat_multi_dim)
        assert out.shape == torch.Size([1, 2, 3, 4, 2])

    # Sparse implementation tests
    def test_linear_sparse_implementation(
        self, feat_input_16_features: torch.Tensor, feat_multi_dim: torch.Tensor
    ):
        layer = PackedLinear(16, 4, alpha=1, num_estimators=1, implementation="sparse")
        out = layer(feat_input_16_features)
        assert out.shape == torch.Size([3, 4])
        layer = PackedLinear(16, 4, alpha=1, num_estimators=2, implementation="sparse")
        out = layer(feat_input_16_features)
        assert out.shape == torch.Size([3, 4])
        layer = PackedLinear(6, 2, alpha=1, num_estimators=1, implementation="sparse")
        out = layer(feat_multi_dim)
        assert out.shape == torch.Size([1, 2, 3, 4, 2])

    # Einsum implementation tests
    def test_linear_einsum_implementation(
        self, feat_input_16_features: torch.Tensor, feat_multi_dim: torch.Tensor
    ):
        layer = PackedLinear(16, 4, alpha=1, num_estimators=1, implementation="einsum")
        out = layer(feat_input_16_features)
        assert out.shape == torch.Size([3, 4])
        layer = PackedLinear(16, 4, alpha=1, num_estimators=2, implementation="einsum")
        out = layer(feat_input_16_features)
        assert out.shape == torch.Size([3, 4])
        layer = PackedLinear(6, 2, alpha=1, num_estimators=1, implementation="einsum")
        out = layer(feat_multi_dim)
        assert out.shape == torch.Size([1, 2, 3, 4, 2])

    def test_linear_extend(self):
        layer = PackedLinear(5, 3, alpha=1, num_estimators=2, gamma=1, implementation="legacy")
        assert layer.weight.shape == torch.Size([4, 3, 1])
        assert layer.bias.shape == torch.Size([4])
        layer = PackedLinear(5, 3, alpha=1, num_estimators=2, gamma=1, implementation="full")
        assert layer.weight.shape == torch.Size([2, 2, 3])
        assert layer.bias.shape == torch.Size([4])
        # with first=True
        layer = PackedLinear(
            5, 3, alpha=1, num_estimators=2, gamma=1, implementation="legacy", first=True
        )
        assert layer.weight.shape == torch.Size([4, 5, 1])
        assert layer.bias.shape == torch.Size([4])
        layer = PackedLinear(
            5, 3, alpha=1, num_estimators=2, gamma=1, implementation="full", first=True
        )
        assert layer.weight.shape == torch.Size([1, 4, 5])
        assert layer.bias.shape == torch.Size([4])

    def test_linear_failures(self):
        with pytest.raises(ValueError):
            _ = PackedLinear(5, 2, alpha=None, num_estimators=1, rearrange=True)

        with pytest.raises(ValueError):
            _ = PackedLinear(5, 2, alpha=-1, num_estimators=1, rearrange=True)

        with pytest.raises(ValueError):
            _ = PackedLinear(5, 2, alpha=1, num_estimators=None, rearrange=True)

        with pytest.raises(TypeError):
            _ = PackedLinear(5, 2, alpha=1, num_estimators=1.5, rearrange=True)

        with pytest.raises(ValueError):
            _ = PackedLinear(5, 2, alpha=1, num_estimators=-1, rearrange=True)

        with pytest.raises(TypeError):
            _ = PackedLinear(5, 2, alpha=1, num_estimators=1, gamma=0.5, rearrange=True)

        with pytest.raises(ValueError):
            _ = PackedLinear(5, 2, alpha=1, num_estimators=1, gamma=-1, rearrange=True)

        with pytest.raises(ValueError):
            _ = PackedLinear(
                5,
                2,
                alpha=1,
                num_estimators=1,
                gamma=1,
                implementation="invalid",
            )

        with pytest.raises(ValueError):
            layer = PackedLinear(16, 4, alpha=1, num_estimators=1, implementation="full")
            layer.implementation = "invalid"
            _ = layer(torch.rand((2, 16)))


class TestPackedConv1d:
    """Testing the PackedConv1d layer class."""

    def test_conv_one_estimator(self, seq_input: torch.Tensor):
        layer = PackedConv1d(6, 2, alpha=1, num_estimators=1, kernel_size=1)
        out = layer(seq_input)
        assert out.shape == torch.Size([5, 2, 3])
        assert layer.weight.shape == torch.Size([2, 6, 1])
        assert layer.bias.shape == torch.Size([2])

    def test_conv_two_estimators(self, seq_input: torch.Tensor):
        layer = PackedConv1d(6, 2, alpha=1, num_estimators=2, kernel_size=1)
        out = layer(seq_input)
        assert out.shape == torch.Size([5, 2, 3])

    def test_conv_one_estimator_gamma2(self, seq_input: torch.Tensor):
        layer = PackedConv1d(6, 2, alpha=1, num_estimators=1, kernel_size=1, gamma=2)
        out = layer(seq_input)
        assert out.shape == torch.Size([5, 2, 3])
        assert layer.conv.groups == 1  # and not 2

    def test_conv_two_estimators_gamma2(self, seq_input: torch.Tensor):
        layer = PackedConv1d(6, 2, alpha=1, num_estimators=2, kernel_size=1, gamma=2)
        out = layer(seq_input)
        assert out.shape == torch.Size([5, 2, 3])
        assert layer.conv.groups == 2  # and not 4

    def test_conv_extend(self):
        _ = PackedConv1d(5, 3, kernel_size=1, alpha=1, num_estimators=2, gamma=1)

    def test_conv1_failures(self):
        with pytest.raises(ValueError):
            _ = PackedConv1d(5, 2, kernel_size=1, alpha=-1, num_estimators=1)

        with pytest.raises(TypeError):
            _ = PackedConv1d(5, 2, kernel_size=1, alpha=1, num_estimators=1, gamma=0.5)

        with pytest.raises(ValueError):
            _ = PackedConv1d(5, 2, kernel_size=1, alpha=1, num_estimators=1, gamma=-1)


class TestPackedConv2d:
    """Testing the PackedConv2d layer class."""

    def test_conv_one_estimator(self, img_input: torch.Tensor):
        layer = PackedConv2d(6, 2, alpha=1, num_estimators=1, kernel_size=1)
        out = layer(img_input)
        assert out.shape == torch.Size([5, 2, 3, 3])

    def test_conv_two_estimators(self, img_input: torch.Tensor):
        layer = PackedConv2d(6, 2, alpha=1, num_estimators=2, kernel_size=1)
        out = layer(img_input)
        assert out.shape == torch.Size([5, 2, 3, 3])

    def test_conv_one_estimator_gamma2(self, img_input: torch.Tensor):
        layer = PackedConv2d(6, 2, alpha=1, num_estimators=1, kernel_size=1, gamma=2)
        out = layer(img_input)
        assert out.shape == torch.Size([5, 2, 3, 3])
        assert layer.conv.groups == 1  # and not 2

    def test_conv_two_estimators_gamma2(self, img_input: torch.Tensor):
        layer = PackedConv2d(6, 2, alpha=1, num_estimators=2, kernel_size=1, gamma=2)
        out = layer(img_input)
        assert out.shape == torch.Size([5, 2, 3, 3])
        assert layer.conv.groups == 2  # and not 4

    def test_conv_extend(self):
        _ = PackedConv2d(5, 3, kernel_size=1, alpha=1, num_estimators=2, gamma=1)

    def test_conv2_failures(self):
        with pytest.raises(ValueError):
            _ = PackedConv2d(5, 2, kernel_size=1, alpha=-1, num_estimators=1)

        with pytest.raises(TypeError):
            _ = PackedConv2d(5, 2, kernel_size=1, alpha=1, num_estimators=1, gamma=0.5)

        with pytest.raises(ValueError):
            _ = PackedConv2d(5, 2, kernel_size=1, alpha=1, num_estimators=1, gamma=-1)


class TestPackedConv3d:
    """Testing the PackedConv3d layer class."""

    def test_conv_one_estimator(self, voxels_input: torch.Tensor):
        layer = PackedConv3d(6, 2, alpha=1, num_estimators=1, kernel_size=1)
        out = layer(voxels_input)
        assert out.shape == torch.Size([5, 2, 3, 3, 3])
        assert layer.weight.shape == torch.Size([2, 6, 1, 1, 1])
        assert layer.bias.shape == torch.Size([2])

    def test_conv_two_estimators(self, voxels_input: torch.Tensor):
        layer = PackedConv3d(6, 2, alpha=1, num_estimators=2, kernel_size=1)
        out = layer(voxels_input)
        assert out.shape == torch.Size([5, 2, 3, 3, 3])

    def test_conv_one_estimator_gamma2(self, voxels_input: torch.Tensor):
        layer = PackedConv3d(6, 2, alpha=1, num_estimators=1, kernel_size=1, gamma=2)
        out = layer(voxels_input)
        assert out.shape == torch.Size([5, 2, 3, 3, 3])
        assert layer.conv.groups == 1  # and not 2

    def test_conv_two_estimators_gamma2(self, voxels_input: torch.Tensor):
        layer = PackedConv3d(6, 2, alpha=1, num_estimators=2, kernel_size=1, gamma=2)
        out = layer(voxels_input)
        assert out.shape == torch.Size([5, 2, 3, 3, 3])
        assert layer.conv.groups == 2  # and not 4

    def test_conv_extend(self):
        _ = PackedConv3d(5, 3, kernel_size=1, alpha=1, num_estimators=2, gamma=1)

    def test_conv3_failures(self):
        with pytest.raises(ValueError):
            _ = PackedConv3d(5, 2, kernel_size=1, alpha=-1, num_estimators=1)

        with pytest.raises(TypeError):
            _ = PackedConv3d(5, 2, kernel_size=1, alpha=1, num_estimators=1, gamma=0.5)

        with pytest.raises(ValueError):
            _ = PackedConv3d(5, 2, kernel_size=1, alpha=1, num_estimators=1, gamma=-1)


class TestPackedLayerNorm:
    """Testing the PackedGroupNorm layer class."""

    def test_one_estimator_forward(self, batched_qkv: torch.Tensor):
        packed_layer_norm = PackedLayerNorm(
            embed_dim=6,
            num_estimators=1,
            alpha=1,
        )
        out = packed_layer_norm(batched_qkv)
        assert out.shape == torch.Size([2, 3, 6])


class TestPackedMultiheadAttention:
    """Testing the PackedMultiheadAttention layer class."""

    def test_one_estimator_qkv(self, unbatched_qkv: torch.Tensor, batched_qkv: torch.Tensor):
        attn_mask = torch.zeros(1, 3, 3, dtype=torch.bool)

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=1,
            alpha=1,
            num_estimators=1,
        )
        out, _ = layer(
            query=unbatched_qkv,
            key=unbatched_qkv,
            value=unbatched_qkv,
            attn_mask=attn_mask,
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_qkv = repeat(unbatched_qkv, "l h -> l b h", b=2)
        attn_mask = torch.zeros(2, 3, 3, dtype=torch.bool)
        out, _ = layer(
            query=unbatched_qkv,
            key=unbatched_qkv,
            value=unbatched_qkv,
            attn_mask=attn_mask,
            is_causal=True,
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=2,
            alpha=1,
            num_estimators=1,
            batch_first=True,
            bias=False,
        )
        out, _ = layer(
            query=batched_qkv,
            key=batched_qkv,
            value=batched_qkv,
        )
        assert out.shape == torch.Size([2, 3, 6])

    def test_one_estimator_q_kv(
        self,
        unbatched_q_kv: tuple[torch.Tensor, torch.Tensor],
        batched_q_kv: tuple[torch.Tensor, torch.Tensor],
    ):
        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=2,
            alpha=1,
            num_estimators=1,
            kdim=2,
            vdim=2,
            add_zero_attn=True,
        )
        out, _ = layer(
            query=unbatched_q_kv[0],
            key=unbatched_q_kv[1],
            value=unbatched_q_kv[1],
        )
        assert out.shape == torch.Size([3, 6])
        unbatched_q_kv = tuple(repeat(seq, "l h -> l b h", b=2) for seq in unbatched_q_kv)
        out, _ = layer(
            query=unbatched_q_kv[0],
            key=unbatched_q_kv[1],
            value=unbatched_q_kv[1],
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=2,
            alpha=1,
            num_estimators=1,
            kdim=2,
            vdim=2,
            batch_first=True,
            bias=False,
        )
        out, _ = layer(
            query=batched_q_kv[0],
            key=batched_q_kv[1],
            value=batched_q_kv[1],
        )
        assert out.shape == torch.Size([2, 3, 6])

    def test_one_estimator_q_k_v(
        self,
        unbatched_q_k_v: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        batched_q_k_v: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    ):
        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=1,
            alpha=1,
            num_estimators=1,
            kdim=2,
            vdim=4,
            add_bias_kv=True,
        )

        key_padding_mask = torch.zeros(4, dtype=torch.bool)

        out, _ = layer(
            query=unbatched_q_k_v[0],
            key=unbatched_q_k_v[1],
            value=unbatched_q_k_v[2],
            key_padding_mask=key_padding_mask,
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_q_k_v = tuple(repeat(seq, "l h -> l b h", b=2) for seq in unbatched_q_k_v)

        out, _ = layer(
            query=unbatched_q_k_v[0],
            key=unbatched_q_k_v[1],
            value=unbatched_q_k_v[2],
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=2,
            alpha=1,
            num_estimators=1,
            kdim=2,
            vdim=4,
            batch_first=True,
        )

        layer.eval()

        attn_mask = torch.zeros(3, 4, dtype=torch.bool)
        key_padding_mask = torch.zeros(2, 4, dtype=torch.bool)

        out, _ = layer(
            query=batched_q_k_v[0],
            key=batched_q_k_v[1],
            value=batched_q_k_v[2],
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
        )
        assert out.shape == torch.Size([2, 3, 6])
        assert out.isfinite().all()

    def test_two_estimators_qkv(self, unbatched_qkv: torch.Tensor, batched_qkv: torch.Tensor):
        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=3,
            alpha=1,
            num_estimators=2,
        )
        out, _ = layer(
            query=unbatched_qkv,
            key=unbatched_qkv,
            value=unbatched_qkv,
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_qkv = repeat(unbatched_qkv, "l h -> l b h", b=2)
        out, _ = layer(
            query=unbatched_qkv,
            key=unbatched_qkv,
            value=unbatched_qkv,
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=3,
            alpha=1,
            num_estimators=2,
            batch_first=True,
        )
        out, _ = layer(
            query=batched_qkv,
            key=batched_qkv,
            value=batched_qkv,
        )
        assert out.shape == torch.Size([2, 3, 6])

    def test_two_estimators_q_kv(
        self,
        unbatched_q_kv: tuple[torch.Tensor, torch.Tensor],
        batched_q_kv: tuple[torch.Tensor, torch.Tensor],
    ):
        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=3,
            alpha=1,
            num_estimators=2,
            kdim=2,
            vdim=2,
            add_zero_attn=True,
        )
        out, _ = layer(
            query=unbatched_q_kv[0],
            key=unbatched_q_kv[1],
            value=unbatched_q_kv[1],
        )
        assert out.shape == torch.Size([3, 6])
        unbatched_q_kv = tuple(repeat(seq, "l h -> l b h", b=2) for seq in unbatched_q_kv)

        attn_mask = torch.zeros(12, 3, 4, dtype=torch.bool)
        key_padding_mask = torch.zeros(2, 4, dtype=torch.bool)

        out, _ = layer(
            query=unbatched_q_kv[0],
            key=unbatched_q_kv[1],
            value=unbatched_q_kv[1],
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=3,
            alpha=1,
            num_estimators=2,
            kdim=2,
            vdim=2,
            batch_first=True,
        )
        out, _ = layer(
            query=batched_q_kv[0],
            key=batched_q_kv[1],
            value=batched_q_kv[1],
        )
        assert out.shape == torch.Size([2, 3, 6])

    def test_two_estimators_q_k_v(
        self,
        unbatched_q_k_v: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        extended_batched_q_k_v: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    ):
        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=3,
            alpha=1,
            num_estimators=2,
            kdim=2,
            vdim=4,
            add_bias_kv=True,
        )
        out, _ = layer(
            query=unbatched_q_k_v[0],
            key=unbatched_q_k_v[1],
            value=unbatched_q_k_v[2],
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_q_k_v = tuple(repeat(seq, "l h -> l b h", b=2) for seq in unbatched_q_k_v)

        attn_mask = torch.zeros(3, 4, dtype=torch.bool)
        key_padding_mask = torch.zeros(2, 4, dtype=torch.bool)

        out, _ = layer(
            query=unbatched_q_k_v[0],
            key=unbatched_q_k_v[1],
            value=unbatched_q_k_v[2],
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedMultiheadAttention(
            embed_dim=6,
            num_heads=3,
            alpha=2,
            num_estimators=2,
            kdim=2,
            vdim=4,
            batch_first=True,
        )
        out, _ = layer(
            query=extended_batched_q_k_v[0],
            key=extended_batched_q_k_v[1],
            value=extended_batched_q_k_v[2],
        )
        assert out.shape == torch.Size([2, 3, 12])


class TestPackedTransformerEncoderLayer:
    """Testing the PackedTransformerEncoderLayer class."""

    def test_one_estimator(self, unbatched_qkv: torch.Tensor, batched_qkv: torch.Tensor):
        layer = PackedTransformerEncoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=2,
            alpha=1,
            num_estimators=1,
            norm_first=True,
            first=True,
        )
        out = layer(
            src=unbatched_qkv,
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_qkv = repeat(unbatched_qkv, "l h -> l b h", b=2)
        out = layer(
            src=unbatched_qkv,
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedTransformerEncoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=2,
            alpha=1,
            num_estimators=1,
            batch_first=True,
            last=True,
            activation=torch.nn.GELU(),
        )
        out = layer(
            src=batched_qkv,
        )
        assert out.shape == torch.Size([2, 3, 6])

    def test_two_estimators(self, unbatched_qkv: torch.Tensor, extended_batched_qkv: torch.Tensor):
        layer = PackedTransformerEncoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=3,
            alpha=1,
            num_estimators=2,
            activation=torch.nn.ELU(),
        )
        out = layer(
            src=unbatched_qkv,
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_qkv = repeat(unbatched_qkv, "l h -> l b h", b=2)
        out = layer(
            src=unbatched_qkv,
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedTransformerEncoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=3,
            alpha=2,
            num_estimators=2,
            batch_first=True,
        )
        out = layer(
            src=extended_batched_qkv,
        )
        assert out.shape == torch.Size([2, 3, 12])


class TestPackedTransformerDecoderLayer:
    """Testing the PackedTransformerDecoderLayer class."""

    def test_one_estimator(
        self,
        unbatched_tgt_memory: tuple[torch.Tensor, torch.Tensor],
        batched_tgt_memory: tuple[torch.Tensor, torch.Tensor],
    ):
        layer = PackedTransformerDecoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=2,
            alpha=1,
            num_estimators=1,
            norm_first=True,
            first=True,
        )
        out = layer(
            tgt=unbatched_tgt_memory[0],
            memory=unbatched_tgt_memory[1],
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_tgt_memory = tuple(
            repeat(seq, "l h -> l b h", b=2) for seq in unbatched_tgt_memory
        )
        out = layer(
            tgt=unbatched_tgt_memory[0],
            memory=unbatched_tgt_memory[1],
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedTransformerDecoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=2,
            alpha=1,
            num_estimators=1,
            batch_first=True,
            last=True,
            activation=torch.nn.GELU(),
            bias=False,
        )
        out = layer(
            tgt=batched_tgt_memory[0],
            memory=batched_tgt_memory[1],
        )
        assert out.shape == torch.Size([2, 3, 6])

    def test_two_estimators(
        self,
        unbatched_tgt_memory: tuple[torch.Tensor, torch.Tensor],
        extended_batched_tgt_memory: tuple[torch.Tensor, torch.Tensor],
    ):
        layer = PackedTransformerDecoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=3,
            alpha=1,
            num_estimators=2,
            activation=torch.nn.ELU(),
        )
        out = layer(
            tgt=unbatched_tgt_memory[0],
            memory=unbatched_tgt_memory[1],
        )
        assert out.shape == torch.Size([3, 6])

        unbatched_tgt_memory = tuple(
            repeat(seq, "l h -> l b h", b=2) for seq in unbatched_tgt_memory
        )
        out = layer(
            tgt=unbatched_tgt_memory[0],
            memory=unbatched_tgt_memory[1],
        )
        assert out.shape == torch.Size([3, 2, 6])

        layer = PackedTransformerDecoderLayer(
            d_model=6,
            dim_feedforward=12,
            nhead=3,
            alpha=2,
            num_estimators=2,
            batch_first=True,
        )
        out = layer(
            tgt=extended_batched_tgt_memory[0],
            memory=extended_batched_tgt_memory[1],
        )
        assert out.shape == torch.Size([2, 3, 12])


class TestPackedFunctional:
    def test_packed_in_projection_packed(
        self,
        batched_qkv: torch.Tensor,
    ):
        proj_q, proj_k, proj_v = packed_in_projection_packed(
            q=batched_qkv,
            k=batched_qkv,
            v=batched_qkv,
            w=torch.rand((1, 18, 6)),
            num_groups=1,
        )
        assert proj_q.shape == torch.Size([2, 3, 6])
        assert proj_k.shape == torch.Size([2, 3, 6])
        assert proj_v.shape == torch.Size([2, 3, 6])

        q_kv = torch.rand((2, 3, 6)), torch.rand((2, 4, 6))

        proj_q, proj_k, proj_v = packed_in_projection_packed(
            q=q_kv[0],
            k=q_kv[1],
            v=q_kv[1],
            w=torch.rand((1, 18, 6)),
            num_groups=1,
            b=None,
        )
        proj_q, proj_k, proj_v = packed_in_projection_packed(
            q=q_kv[0],
            k=q_kv[1],
            v=q_kv[1],
            w=torch.rand((1, 18, 6)),
            num_groups=1,
            b=torch.rand(18),
        )

        assert proj_q.shape == torch.Size([2, 3, 6])
        assert proj_k.shape == torch.Size([2, 4, 6])
        assert proj_v.shape == torch.Size([2, 4, 6])

        q_k_v = torch.rand((2, 3, 6)), torch.rand((2, 4, 6)), torch.rand((2, 4, 6))

        proj_q, proj_k, proj_v = packed_in_projection_packed(
            q=q_k_v[0],
            k=q_k_v[1],
            v=q_k_v[2],
            w=torch.rand((1, 18, 6)),
            num_groups=1,
            b=None,
        )

        proj_q, proj_k, proj_v = packed_in_projection_packed(
            q=q_k_v[0],
            k=q_k_v[1],
            v=q_k_v[2],
            w=torch.rand((1, 18, 6)),
            num_groups=1,
            b=torch.rand(18),
        )

        assert proj_q.shape == torch.Size([2, 3, 6])
        assert proj_k.shape == torch.Size([2, 4, 6])
        assert proj_v.shape == torch.Size([2, 4, 6])

    def test_packed_multi_head_attention_forward_failures(self, unbatched_q_k_v: torch.Tensor):
        q, k, v = unbatched_q_k_v
        with pytest.raises(RuntimeError):
            _ = packed_multi_head_attention_forward(
                query=q,
                key=k,
                value=v,
                embed_dim_to_check=6,
                num_heads=2,
                num_groups=1,
                in_proj_weight=None,
                in_proj_bias=torch.rand(18),
                bias_k=None,
                bias_v=None,
                add_zero_attn=False,
                dropout_p=0.0,
                out_proj_weight=torch.rand(1, 6, 6),
                out_proj_bias=None,
                is_causal=True,
                attn_mask=None,
            )

        with pytest.raises(RuntimeError):
            _ = packed_multi_head_attention_forward(
                query=q,
                key=k,
                value=v,
                embed_dim_to_check=6,
                num_heads=2,
                num_groups=1,
                in_proj_weight=None,
                in_proj_bias=torch.rand(18),
                bias_k=None,
                bias_v=None,
                add_zero_attn=False,
                dropout_p=0.0,
                out_proj_weight=torch.rand(1, 6, 6),
                out_proj_bias=None,
                attn_mask=torch.rand(2, 2),
                use_separate_proj_weight=True,
                q_proj_weight=torch.rand(1, 6, 6),
                k_proj_weight=torch.rand(1, 6, 2),
                v_proj_weight=torch.rand(1, 6, 4),
            )

        with pytest.raises(AssertionError):
            _ = packed_multi_head_attention_forward(
                query=q,
                key=k,
                value=v,
                embed_dim_to_check=6,
                num_heads=2,
                num_groups=1,
                in_proj_weight=None,
                in_proj_bias=torch.rand(18),
                bias_k=None,
                bias_v=None,
                add_zero_attn=False,
                dropout_p=0.0,
                out_proj_weight=torch.rand(1, 6, 6),
                out_proj_bias=None,
                attn_mask=torch.rand(1, 1, 3, 4),
                use_separate_proj_weight=True,
                q_proj_weight=torch.rand(1, 6, 6),
                k_proj_weight=torch.rand(1, 6, 2),
                v_proj_weight=torch.rand(1, 6, 4),
            )

        with pytest.raises(AssertionError):
            _ = packed_multi_head_attention_forward(
                query=q,
                key=k,
                value=v,
                embed_dim_to_check=6,
                num_heads=2,
                num_groups=1,
                in_proj_weight=None,
                in_proj_bias=torch.rand(18),
                bias_k=None,
                bias_v=None,
                add_zero_attn=False,
                dropout_p=0.0,
                out_proj_weight=torch.rand(1, 6, 6),
                out_proj_bias=None,
                attn_mask=torch.rand(1, 2, 2),
                use_separate_proj_weight=True,
                q_proj_weight=torch.rand(1, 6, 6),
                k_proj_weight=torch.rand(1, 6, 2),
                v_proj_weight=torch.rand(1, 6, 4),
            )



from collections.abc import Iterable
from pathlib import Path
from typing import Literal

from lightning.pytorch import Trainer
from lightning.pytorch.accelerators.accelerator import Accelerator
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.loggers import Logger
from lightning.pytorch.profilers import Profiler
from lightning.pytorch.strategies.strategy import Strategy
from lightning.pytorch.trainer.states import (
    RunningStage,
    TrainerFn,
)

from torch_uncertainty.utils.evaluation_loop import TUEvaluationLoop


class TUTrainer(Trainer):
    def __init__(
        self,
        accelerator: str | Accelerator = "auto",
        strategy: str | Strategy = "auto",
        devices: list[int] | str | int = "auto",
        num_nodes: int = 1,
        precision: None | int | str = None,
        logger: Logger | Iterable[Logger] | bool | None = None,
        callbacks: list[Callback] | Callback | None = None,
        fast_dev_run: int | bool = False,
        max_epochs: int | None = None,
        min_epochs: int | None = None,
        max_steps: int = -1,
        min_steps: int | None = None,
        max_time: str | dict[str, int] | None = None,
        limit_train_batches: float | None = None,
        limit_val_batches: float | None = None,
        limit_test_batches: float | None = None,
        limit_predict_batches: float | None = None,
        overfit_batches: float = 0,
        val_check_interval: float | None = None,
        check_val_every_n_epoch: int | None = 1,
        num_sanity_val_steps: int | None = None,
        log_every_n_steps: int | None = None,
        enable_checkpointing: bool | None = None,
        enable_progress_bar: bool | None = None,
        enable_model_summary: bool | None = None,
        accumulate_grad_batches: int = 1,
        gradient_clip_val: float | None = None,
        gradient_clip_algorithm: str | None = None,
        deterministic: bool | None | Literal["warn"] = None,
        benchmark: bool | None = None,
        inference_mode: bool = True,
        use_distributed_sampler: bool = True,
        profiler: Profiler | str | None = None,
        detect_anomaly: bool = False,
        barebones: bool = False,
        plugins=None,
        sync_batchnorm: bool = False,
        reload_dataloaders_every_n_epochs: int = 0,
        default_root_dir: str | Path | None = None,
    ) -> None:
        super().__init__(
            accelerator=accelerator,
            strategy=strategy,
            devices=devices,
            num_nodes=num_nodes,
            precision=precision,
            logger=logger,
            callbacks=callbacks,
            fast_dev_run=fast_dev_run,
            max_epochs=max_epochs,
            min_epochs=min_epochs,
            max_steps=max_steps,
            min_steps=min_steps,
            max_time=max_time,
            limit_train_batches=limit_train_batches,
            limit_val_batches=limit_val_batches,
            limit_test_batches=limit_test_batches,
            limit_predict_batches=limit_predict_batches,
            overfit_batches=overfit_batches,
            val_check_interval=val_check_interval,
            check_val_every_n_epoch=check_val_every_n_epoch,
            num_sanity_val_steps=num_sanity_val_steps,
            log_every_n_steps=log_every_n_steps,
            enable_checkpointing=enable_checkpointing,
            enable_progress_bar=enable_progress_bar,
            enable_model_summary=enable_model_summary,
            accumulate_grad_batches=accumulate_grad_batches,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            deterministic=deterministic,
            benchmark=benchmark,
            inference_mode=inference_mode,
            use_distributed_sampler=use_distributed_sampler,
            profiler=profiler,
            detect_anomaly=detect_anomaly,
            barebones=barebones,
            plugins=plugins,
            sync_batchnorm=sync_batchnorm,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
            default_root_dir=default_root_dir,
        )
        self.test_loop = TUEvaluationLoop(
            self,
            TrainerFn.TESTING,
            RunningStage.TESTING,
            inference_mode=inference_mode,
        )



from typing import Literal

import torch
from torch import Tensor, nn

from .scaler import Scaler


class TemperatureScaler(Scaler):
    def __init__(
        self,
        model: nn.Module | None = None,
        init_val: float = 1,
        lr: float = 0.1,
        max_iter: int = 100,
        device: Literal["cpu", "cuda"] | torch.device | None = None,
    ) -> None:
        """Temperature scaling post-processing for calibrated probabilities.

        Args:
            model (nn.Module): Model to calibrate.
            init_val (float, optional): Initial value for the temperature.
                Defaults to 1.
            lr (float, optional): Learning rate for the optimizer. Defaults to 0.1.
            max_iter (int, optional): Maximum number of iterations for the
                optimizer. Defaults to 100.
            device (Optional[Literal["cpu", "cuda"]], optional): Device to use
                for optimization. Defaults to None.

        Reference:
            Guo, C., Pleiss, G., Sun, Y., & Weinberger, K. Q. On calibration
            of modern neural networks. In ICML 2017.
        """
        super().__init__(model=model, lr=lr, max_iter=max_iter, device=device)

        if init_val <= 0:
            raise ValueError("Initial temperature value must be positive.")

        self.set_temperature(init_val)

    def set_temperature(self, val: float) -> None:
        """Set the temperature to a fixed value.

        Args:
            val (float): Temperature value.
        """
        if val <= 0:
            raise ValueError("Temperature value must be positive.")

        self.temp = nn.Parameter(torch.ones(1, device=self.device) * val, requires_grad=True)

    def _scale(self, logits: Tensor) -> Tensor:
        """Scale the prediction with the optimal temperature.

        Args:
            logits (Tensor): logits to be scaled.

        Returns:
            Tensor: Scaled logits.
        """
        return logits / self.temperature[0]

    @property
    def temperature(self) -> list:
        return [self.temp]



from typing import Literal

import torch
from torch import Tensor, nn
from torch.distributions import Distribution, Independent

from torch_uncertainty.utils.distributions import NormalInverseGamma


class DistributionNLLLoss(nn.Module):
    def __init__(self, reduction: Literal["mean", "sum"] | None = "mean") -> None:
        """Negative Log-Likelihood loss using given distributions as inputs.

        Args:
            reduction (str, optional): specifies the reduction to apply to the
            output:``'none'`` | ``'mean'`` | ``'sum'``. Defaults to "mean".
        """
        super().__init__()
        self.reduction = reduction

    def forward(
        self,
        dist: Distribution,
        targets: Tensor,
        padding_mask: Tensor | None = None,
    ) -> Tensor:
        """Compute the NLL of the targets given predicted distributions.

        Args:
            dist (Distribution): The predicted distributions
            targets (Tensor): The target values
            padding_mask (Tensor, optional): The padding mask. Defaults to None.
                Sets the loss to 0 for padded values.
        """
        loss = -dist.log_prob(targets)
        if padding_mask is not None:
            loss = loss.masked_fill(padding_mask, float("nan"))

        if self.reduction == "mean":
            loss = loss.nanmean()
        elif self.reduction == "sum":
            loss = loss.nansum()
        return loss


class DERLoss(DistributionNLLLoss):
    def __init__(self, reg_weight: float, reduction: str | None = "mean") -> None:
        """The Deep Evidential Regression loss.

        This loss combines the negative log-likelihood loss of the normal
        inverse gamma distribution and a weighted regularization term.

        Args:
            reg_weight (float): The weight of the regularization term.
            reduction (str, optional): specifies the reduction to apply to the
            output:``'none'`` | ``'mean'`` | ``'sum'``.

        Reference:
            Amini, A., Schwarting, W., Soleimany, A., & Rus, D. (2019). Deep
            evidential regression. https://arxiv.org/abs/1910.02600.
        """
        super().__init__(reduction=None)

        if reduction not in ("none", "mean", "sum") and reduction is not None:
            raise ValueError(f"{reduction} is not a valid value for reduction.")
        self.der_reduction = reduction

        if reg_weight < 0:
            raise ValueError(
                f"The regularization weight should be non-negative, but got {reg_weight}."
            )
        self.reg_weight = reg_weight

    def _reg(self, dist: NormalInverseGamma | Independent, targets: Tensor) -> Tensor:
        if isinstance(dist, Independent):
            dist = dist.base_dist

        return torch.norm(targets - dist.loc, 1, dim=1, keepdim=True) * (
            2 * dist.lmbda + dist.alpha
        )

    def forward(
        self,
        dist: NormalInverseGamma,
        targets: Tensor,
    ) -> Tensor:
        loss_nll = super().forward(dist, targets)
        loss_reg = self._reg(dist, targets)
        loss = loss_nll + self.reg_weight * loss_reg

        if self.der_reduction == "mean":
            return loss.mean()
        if self.der_reduction == "sum":
            return loss.sum()
        return loss


class BetaNLL(nn.Module):
    def __init__(self, beta: float = 0.5, reduction: str | None = "mean") -> None:
        """The Beta Negative Log-likelihood loss.

        Args:
            beta (float): TParameter from range [0, 1] controlling relative
            weighting between data points, where `0` corresponds to
            high weight on low error points and `1` to an equal weighting.
            reduction (str, optional): specifies the reduction to apply to the
            output:``'none'`` | ``'mean'`` | ``'sum'``.

        Reference:
            Seitzer, M., Tavakoli, A., Antic, D., & Martius, G. (2022). On the
            pitfalls of heteroscedastic uncertainty estimation with probabilistic
            neural networks. https://arxiv.org/abs/2203.09168.
        """
        super().__init__()

        if beta < 0 or beta > 1:
            raise ValueError(f"The beta parameter should be in range [0, 1], but got {beta}.")
        self.beta = beta
        self.nll_loss = nn.GaussianNLLLoss(reduction="none")
        if reduction not in ("none", "mean", "sum"):
            raise ValueError(f"{reduction} is not a valid value for reduction.")
        self.reduction = reduction

    def forward(self, mean: Tensor, targets: Tensor, variance: Tensor) -> Tensor:
        loss = self.nll_loss(mean, targets, variance) * (variance.detach() ** self.beta)

        if self.reduction == "mean":
            return loss.mean()
        if self.reduction == "sum":
            return loss.sum()
        return loss



from typing import Literal

from torch import nn

from torch_uncertainty.models.segmentation.deeplab import (
    deep_lab_v3_resnet50,
    deep_lab_v3_resnet101,
)
from torch_uncertainty.routines.segmentation import SegmentationRoutine


class DeepLabBaseline(SegmentationRoutine):
    single = ["std"]
    versions = {
        "std": [
            deep_lab_v3_resnet50,
            deep_lab_v3_resnet101,
        ]
    }
    archs = [50, 101]

    def __init__(
        self,
        num_classes: int,
        loss: nn.Module,
        version: Literal["std"],
        arch: int,
        style: Literal["v3", "v3+"],
        output_stride: int,
        separable: bool,
        metric_subsampling_rate: float = 1e-2,
        log_plots: bool = False,
        num_calibration_bins: int = 15,
        pretrained_backbone: bool = True,
    ) -> None:
        params = {
            "num_classes": num_classes,
            "style": style,
            "output_stride": output_stride,
            "separable": separable,
            "pretrained_backbone": pretrained_backbone,
        }

        format_batch_fn = nn.Identity()

        if version not in self.versions:
            raise ValueError(f"Unknown version {version}")

        model = self.versions[version][self.archs.index(arch)](**params)
        super().__init__(
            num_classes=num_classes,
            model=model,
            loss=loss,
            format_batch_fn=format_batch_fn,
            metric_subsampling_rate=metric_subsampling_rate,
            log_plots=log_plots,
            num_calibration_bins=num_calibration_bins,
        )
        self.save_hyperparameters(ignore=["loss"])



"""
Deep Probabilistic Regression
=============================

This tutorial aims to provide an overview of some utilities in TorchUncertainty for probabilistic regression.

Building a MLP for Probabilistic Regression using TorchUncertainty distribution layers
--------------------------------------------------------------------------------------

In this section we cover the building of a very simple MLP outputting Normal distribution parameters.

1. Loading the utilities
~~~~~~~~~~~~~~~~~~~~~~~~

We disable some logging and warnings to keep the output clean.
"""
# %%
import torch
from torch import nn

import logging
logging.getLogger("lightning.pytorch.utilities.rank_zero").setLevel(logging.WARNING)

import warnings
warnings.filterwarnings("ignore")

# %%
# 2. Building the MLP model
# ~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To create a MLP model estimating a Normal distribution, we use the NormalLinear layer.
# This layer is a wrapper around the nn.Linear layer, which outputs the location and scale of a Normal distribution.
# Note that any other distribution layer from TU can be used in the same way.
from torch_uncertainty.layers.distributions import NormalLinear


class MLP(nn.Module):
    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self.fc1 = nn.Linear(in_features, 50)
        self.fc2 = NormalLinear(
            base_layer=nn.Linear,
            event_dim=out_features,
            in_features=50,
        )

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        return self.fc2(x)

# %%
# 3. Setting up the data
# ~~~~~~~~~~~~~~~~~~~~~~
#
# We use the UCI Kin8nm dataset, which is a regression dataset with 8 features and 8192 samples.
from torch_uncertainty.datamodules import UCIRegressionDataModule

# datamodule
datamodule = UCIRegressionDataModule(
    root="data",
    batch_size=32,
    dataset_name="kin8nm",
)

# %%
# 4. Setting up the model and trainer
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

from torch_uncertainty import TUTrainer

trainer = TUTrainer(
    accelerator="cpu",
    max_epochs=5,
    enable_progress_bar=False,
)

model = MLP(in_features=8, out_features=1)


# %%
# 5. The Loss, the Optimizer and the Training Routine
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We use the DistributionNLLLoss to compute the negative log-likelihood of the Normal distribution.
# Note that this loss can be used with any Distribution from torch.distributions.
# For the optimizer, we use the Adam optimizer with a learning rate of 5e-3.
# Finally, we create a RegressionRoutine to train the model. We indicate that the output dimension is 1 and the distribution family is "normal".

from torch_uncertainty.losses import DistributionNLLLoss
from torch_uncertainty.routines import RegressionRoutine

loss = DistributionNLLLoss()

def optim_regression(
    model: nn.Module,
    learning_rate: float = 5e-3,
):
    return torch.optim.Adam(
        model.parameters(),
        lr=learning_rate,
        weight_decay=0,
    )

routine = RegressionRoutine(
    output_dim=1,
    model=model,
    loss=loss,
    optim_recipe=optim_regression(model),
    dist_family="normal",
)


# %%
# 6. Training the model
# ~~~~~~~~~~~~~~~~~~~~~~

trainer.fit(model=routine, datamodule=datamodule)
results = trainer.test(model=routine, datamodule=datamodule)

# %%
# 7. Benchmarking different distributions
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Our MLP model assumes a Normal distribution as the output. However, we could be interested in comparing the performance of different distributions.
# TorchUncertainty provides a simple way to do this using the get_dist_linear_layer() function.
# Let us rewrite the MLP model to use it.

from torch_uncertainty.layers.distributions import get_dist_linear_layer

class MLP(nn.Module):
    def __init__(self, in_features: int, out_features: int, dist_family: str):
        super().__init__()
        self.fc1 = nn.Linear(in_features, 50)
        dist_layer = get_dist_linear_layer(dist_family)
        self.fc2 = dist_layer(
            base_layer=nn.Linear,
            event_dim=out_features,
            in_features=50,
        )

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        return self.fc2(x)

# %%
# We can now train the model with different distributions.
# Let us train the model with a Normal, Laplace, Student's t, and Cauchy distribution.
# Note that we use the mode as the point-wise estimate of the distribution as the mean
# is not defined for the Cauchy distribution.
for dist_family in ["normal", "laplace", "student", "cauchy"]:
    print("#" * 50)
    print(f">>> Training with {dist_family} distribution")
    print("#" * 50)
    trainer = TUTrainer(
        accelerator="cpu",
        max_epochs=10,
        enable_model_summary=False,
        enable_progress_bar=False,
    )
    model = MLP(in_features=8, out_features=1, dist_family=dist_family)
    routine = RegressionRoutine(
        output_dim=1,
        model=model,
        loss=loss,
        optim_recipe=optim_regression(model),
        dist_family=dist_family,
        dist_estimate="mode",
    )
    trainer.fit(model=routine, datamodule=datamodule)
    trainer.test(model=routine, datamodule=datamodule)



import torch
from torch import Tensor, nn
from torch.nn import functional as F


class BatchNormAdapter2d(nn.Module):
    def __init__(
        self,
        num_features: int,
        alpha: float = 0.1,
        momentum: float = 0.1,
        eps: float = 1e-5,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> None:
        super().__init__()
        self.weight = nn.Parameter(
            torch.ones(num_features, device=device, dtype=dtype),
            requires_grad=True,
        )
        self.bias = nn.Parameter(
            torch.zeros(num_features, device=device, dtype=dtype),
            requires_grad=True,
        )

        self.register_buffer(
            "running_mean",
            torch.zeros(num_features, device=device, dtype=dtype),
        )
        self.register_buffer(
            "running_var",
            torch.zeros(num_features, device=device, dtype=dtype),
        )
        self.register_buffer(
            "num_batches_tracked",
            torch.tensor(0, dtype=torch.long, device=device),
        )
        self.alpha = alpha
        self.momentum = momentum
        self.eps = eps
        self.frozen = False

    def forward(self, x: Tensor) -> Tensor:
        if self.frozen:
            return F.batch_norm(
                x,
                self.running_mean,
                self.running_var,
                self.weight,
                self.bias,
                self.training,
                self.momentum,
                self.eps,
            )
        out = F.batch_norm(
            x,
            self.running_mean,
            self.running_var,
            None,
            None,
            self.training,
            self.momentum,
            self.eps,
        )
        return self.weight.unsqueeze(-1).unsqueeze(-1) * out * (
            torch.randn_like(x) * self.alpha + 1
        ) + self.bias.unsqueeze(-1).unsqueeze(-1)



# ruff: noqa: F401
from .abstract import PostProcessing
from .calibration import MatrixScaler, TemperatureScaler, VectorScaler
from .laplace import LaplaceApprox
from .mc_batch_norm import MCBatchNorm



from pathlib import Path

import pytest
import torch
from torch import nn

from tests._dummies import (
    DummyPixelRegressionBaseline,
    DummyPixelRegressionDataModule,
)
from torch_uncertainty import TUTrainer
from torch_uncertainty.losses import DistributionNLLLoss
from torch_uncertainty.optim_recipes import optim_cifar10_resnet18
from torch_uncertainty.routines.pixel_regression import (
    PixelRegressionRoutine,
    colorize,
)


class TestPixelRegression:
    def test_one_estimator_two_classes(self):
        trainer = TUTrainer(
            accelerator="cpu",
            max_epochs=1,
            logger=None,
            enable_checkpointing=False,
        )

        root = Path(__file__).parent.absolute().parents[0] / "data"
        dm = DummyPixelRegressionDataModule(root=root, batch_size=5, output_dim=3)

        model = DummyPixelRegressionBaseline(
            dist_family=None,
            in_channels=dm.num_channels,
            output_dim=dm.output_dim,
            image_size=dm.image_size,
            loss=nn.MSELoss(),
            baseline_type="single",
            optim_recipe=optim_cifar10_resnet18,
            ema=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

        trainer = TUTrainer(
            accelerator="cpu",
            max_epochs=1,
            logger=None,
            enable_checkpointing=False,
        )
        model = DummyPixelRegressionBaseline(
            dist_family="normal",
            in_channels=dm.num_channels,
            output_dim=dm.output_dim,
            image_size=dm.image_size,
            loss=DistributionNLLLoss(),
            baseline_type="single",
            optim_recipe=optim_cifar10_resnet18,
            swa=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_two_estimators_one_class(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        root = Path(__file__).parent.absolute().parents[0] / "data"
        dm = DummyPixelRegressionDataModule(root=root, batch_size=4, output_dim=1)

        model = DummyPixelRegressionBaseline(
            dist_family=None,
            in_channels=dm.num_channels,
            output_dim=dm.output_dim,
            image_size=dm.image_size,
            loss=nn.MSELoss(),
            baseline_type="ensemble",
            optim_recipe=optim_cifar10_resnet18,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True, logger=None)
        model = DummyPixelRegressionBaseline(
            dist_family="normal",
            in_channels=dm.num_channels,
            output_dim=dm.output_dim,
            image_size=dm.image_size,
            loss=DistributionNLLLoss(),
            baseline_type="ensemble",
            optim_recipe=optim_cifar10_resnet18,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

        colorize(torch.ones((10, 10)), 0, 1)
        colorize(torch.ones((10, 10)), 0, 0)

    def test_depth_errors(self):
        with pytest.raises(ValueError, match="output_dim must be positive"):
            PixelRegressionRoutine(
                model=nn.Identity(),
                output_dim=0,
                loss=nn.MSELoss(),
                dist_family=None,
            )

        with pytest.raises(ValueError, match="num_image_plot must be positive"):
            PixelRegressionRoutine(
                model=nn.Identity(),
                output_dim=1,
                loss=nn.MSELoss(),
                dist_family=None,
                num_image_plot=0,
                log_plots=True,
            )



import pytest
import torch

from torch_uncertainty.metrics import GroupingLoss


class TestGroupingLoss:
    """Testing the GroupingLoss metric class."""

    def test_compute(self):
        metric = GroupingLoss()
        metric.update(
            torch.cat([torch.tensor([0, 1, 0, 1]), torch.ones(200) / 10]),
            torch.cat([torch.tensor([0, 0, 1, 1]), torch.zeros(100), torch.ones(100)]).long(),
            torch.cat([torch.zeros((104, 10)), torch.ones((100, 10))]),
        )
        metric.compute()

        metric = GroupingLoss()
        metric.update(
            torch.ones((200, 4, 10)),
            torch.cat([torch.arange(100), torch.arange(100)]),
            torch.cat([torch.zeros((100, 4, 10)), torch.ones((100, 4, 10))]),
        )
        metric.compute()
        metric.reset()

        metric.update(
            torch.ones((200, 10)) / 10,
            torch.nn.functional.one_hot(torch.arange(200)),
            torch.cat([torch.zeros((100, 10)), torch.ones((1004, 10))]),
        )

    def test_errors(self):
        metric = GroupingLoss()
        with pytest.raises(ValueError):
            metric.update(
                torch.ones((2, 10, 1, 1)) / 10,
                torch.arange(2),
                torch.rand((2, 10)),
            )

        with pytest.raises(ValueError):
            metric.update(
                torch.ones((2, 10)) / 10,
                torch.arange(2),
                torch.rand((2, 10, 10, 1, 1)),
            )

        with pytest.raises(ValueError):
            metric.update(
                torch.ones((2, 10)) / 10,
                torch.arange(2).unsqueeze(-1).unsqueeze(-1),
                torch.rand((2, 10, 10, 1, 1)),
            )
